//============================================================================
//
//                              Program/Module
//                                   from
//                C++ package for verified numerical integration
//                  of singular and strong singular functions
//                      Copyright (c) 2000 Stefan Wedner
//
// This program/module is free software for non-commercial use.
//
//
//============================================================================
//----------------------------------------------------------------------------
//
// File:     cubature.h
// See also: cubature.C
// Purpose:  computing enclosures of one-dimensional integrals
//     
// ===========================================================================


#include <algorithm>
#include <iterator>
#include <iostream>
#include "interval.hpp"
#include "./taylor/f_ari.h"

#define CUBATURE

class base_interval;
class pbase_interval;
class xriemann;
class riemann;
class xcauchy;
class cauchyxriemann;
class cauchy;


// ----------------------------------------------------------------------------
// --- class integral ---------------------------------------------------------
// ----------------------------------------------------------------------------

class integral{

public:
  integral();
  integral(integrand&);
  ~integral();
  interval integrate(real epsabs);
  interval integrate(real a, real b, real c, real d, real epsabs);
  interval integrate(real a, real b, real c, interval d, real epsabs){
    integrate(interval(a), interval(b),  interval(c), d , epsabs);
  } 
  //  and so on
  interval integrate(interval a, interval b, interval c, interval d, 
                                                                real epsabs);

  friend ostream& operator<<(ostream& os, const integral& out);

  void gaa();

  void print_division();
  void print_gnuplot();

protected:

  real xinf, xsup, yinf, ysup;
  integrand f, ftransposed;  
  real eps_abs, eps_rel;
  

  dotprecision Rinf, Rsup;
  idotprecision Approximation;

  static int integral_counter;
  static const int max_heapsize;  // heap: administration of subintervals 
  
  int fcounter;
  int heapsize;
  int divided_intervals;

  pbase_interval* heap;

  // functions:
  base_interval* pop();
  void push(base_interval& r_i);
  void print_heap();
  static real epsilon; 

 private: 

  virtual void start_division();

  friend class xriemann; 
  friend class riemann;
};



// ----------------------------------------------------------------------------
// --- class cauchyxriemann_integral ------------------------------------------
// ----------------------------------------------------------------------------


class cauchyxriemann_integral: public integral{

 public:
  cauchyxriemann_integral();
  cauchyxriemann_integral(integrand& g, real xl);
  ~cauchyxriemann_integral();
  interval integrate(interval, interval, interval, interval, real);
  interval integrate(real a, real b, real c, real d, real eps){
      integral::integrate(a, b, c, d, eps);
  } //workaround
 protected:
  real xlambda; // collocation point
  static int cauchy_counter;

  integrand f_c, f_ctransposed; 
 
 private: 
  void start_division();
  friend class xcauchy;
  friend class cauchyxriemann;
};



// ----------------------------------------------------------------------------
// --- class cauchy_integral --------------------------------------------------
// ----------------------------------------------------------------------------


class cauchy_integral: public cauchyxriemann_integral{

 public:
  cauchy_integral(integrand& g, real xl, real yl);
  interval integrate(interval, interval, interval, interval, real);
  interval integrate(real a, real b, real c, real d, real eps){
      integral::integrate(a, b, c, d, eps);
  } //workaround
 protected:
  real ylambda; // collocation point
 
 private: 
  integrand f_cc;
  void start_division();
  friend class xcauchy;
  friend class cauchy;
  friend class cauchyxriemann;
};



// ----------------------------------------------------------------------------
// --- class base_interval ----------------------------------------------------
// ----------------------------------------------------------------------------


class base_interval{
 public:
  base_interval(real a, real b, real c, real d, int k, int l, 
                integral& r_i, bool trans = false):
     xinf(a), xsup(b), yinf(c), ysup(d),
     xm_index(k), xn_index(l), approx(false), transposed(trans), ri(r_i),
     yn(1), yn_index(0), ym(0), ym_index(0){ }

  //  virtual ~base_interval();

  real dremainder;

 protected:
  real xinf, xsup, yinf, ysup;
  real r_inf, r_sup;
  interval approximation;

  int  xn_index, xm_index, xn, xm; 
  int  yn_index, ym_index, yn, ym;
 
  bool approx;
  bool transposed;
  integral& ri;

  static dotprecision r_dot;

  virtual void add_approxsum() = 0;
  virtual void subdivide() = 0;

  virtual void print_interval(ostream& os){ 
        os  << " intervaltype: unknown " << *this << endl;}

   // base_interval(const base_interval& t);
  //  base_interval& operator=(const base_interval&);

  friend ostream& operator<<(ostream& os, const base_interval& out);
  friend istream& operator>>(istream& is, base_interval& in);
  friend class integral;
};



// ----------------------------------------------------------------------------
// --- class pbase_interval ---------------------------------------------------
// ----------------------------------------------------------------------------


class pbase_interval{
 
 public:
  pbase_interval(){ pointer = 0;}
  pbase_interval(base_interval& r_i){ pointer = &r_i; }
  bool operator<(const pbase_interval&) const;
  bool operator>(const pbase_interval&) const;
  bool operator==(const pbase_interval&) const;
  friend ostream& operator<<(ostream& os, const pbase_interval& out);
  friend istream& operator>>(istream& is, pbase_interval& in);
  base_interval* pointer;
};



// ----------------------------------------------------------------------------
// --- class xriemann ---------------------------------------------------------
// ----------------------------------------------------------------------------
// intervaltype for one-dimensional integration (bound)
// ----------------------------------------------------------------------------



class xriemann: virtual public base_interval{
 
public:
 xriemann(real a, real b, real c, real d, int k, int l, integral& r_i, 
          bool t = false): base_interval(a, b, c, d, k, l, r_i, t){
    xm = m_index2m(xm_index);
    xn = n_index2n(xn_index);
    yn = 1; yn_index = 0;
    ym = 0; ym_index = 0;
  }

   xriemann(real a, real b, real c, real d, int k, int l, integral& r_i, 
          real eps, bool t = false): base_interval(a, b, c, d, k, l, r_i, t){
    xm = m_index2m(xm_index);
    xn = n_index2n(xn_index);
    yn = 1; yn_index = 0;
    ym = 0; ym_index = 0;
    choose_qf((b-a) * eps);
    r_i.push(*this);    
  }

   xriemann(real a, real b, real c, real d, integral& r_i)
                               : base_interval(a, b, c, d, 0, 0, r_i, false){
    xm = 0; 
    xn = 1;
    yn = 1; yn_index = 0;
    ym = 0; ym_index = 0;
    approximation = (b - interval(a,b)) * (d - interval(c,d)) *
                         r_i.f.value(interval(a,b),interval(c,d));
    approx = true;
    r_inf = r_sup  = 0;
    dremainder = 0;
    r_i.push(*this);
  }

  static interval **node, **weight, **peano;
  static int  *peano_n;  // cubature formulas
  static const int n_min, n_max, n_dim; 
  static const int m_min, m_max, m_dim;
  static const int m_stride, n_stride;

  void set_xriemann(interval, int, int);
  void set_xriemann(int, int, interval);

  void add_approxsum();
  void subdivide();
  void choose_qf(real eps_local);
  void print_interval(ostream& os){
    os << " intervaltype: riemannx " << *this << endl;
  }

  static int n2n_index(int k){ return (k - n_min) /  n_stride; }  
  static int n_index2n(int k){ return k * n_stride + n_min; }
  static int m2m_index(int k){ return (k - m_min) /  m_stride; }  
  static int m_index2m(int k){ return k * m_stride + m_min; }
  static interval** init_data();
  static void free_data();
};



// ----------------------------------------------------------------------------
// --- class riemann ----------------------------------------------------------
// ----------------------------------------------------------------------------
//  two-dimensional integration range
// ----------------------------------------------------------------------------


class riemann:public xriemann{
 
 public:
  riemann(real a, real b, real c, real d, int k1, int l1, int k2, int l2,
          integral& r_i, bool xdiv = true, bool ydiv = true)
         :  base_interval(a, b, c, d, k1, l1, r_i),
            xriemann(a , b, c, d, k1, l1, r_i)
              
 {
    ym_index = k2;
    yn_index = l2;

    ym = m_index2m(ym_index);
    yn = n_index2n(yn_index);
    
    xdivide = xdiv;
    ydivide = ydiv;
  }

  riemann(real a, real b, real c, real d, int k1, int l1, int k2, int l2,
          integral& r_i, bool xdiv, bool ydiv, real eps);

  void set_xriemann(interval, int, int);
  void set_xriemann(int, int, interval);
  void set_yriemann(interval, int, int);
  void set_yriemann(int, int, interval);

  void add_approxsum();
  void subdivide();
  void choose_xcf(real eps_local);
  void choose_ycf(real eps_local);

  void print_interval(ostream& os){
    os << " intervaltype: riemann " << *this << endl;}
 private:
  bool xdivide, ydivide;

} ;  



// ----------------------------------------------------------------------------
// --- class xcauchy ----------------------------------------------------------
// ----------------------------------------------------------------------------


class xcauchy: virtual public base_interval{

 public:

  xcauchy(real a, real b, real c, real d, int k1, int l1, 
          cauchyxriemann_integral& c_i, bool t = false)
                  : base_interval(a, b, c, d, k1, l1, c_i, t){
    xm       = m_index2m(xm_index);
    xn       = xriemann::n_index2n(xn_index);
    ym_index = 0;
    yn_index = 0;
    ym       = 0;
    yn       = 1;    
  }
  
   xcauchy(real a, real b, real c, real d, int k1, int l1, 
          cauchyxriemann_integral& c_i, real eps, bool t = false)
           : base_interval(a, b, c, d, k1, l1, c_i, t){
    xm       = m_index2m(xm_index);
    xn       = xriemann::n_index2n(xn_index);
    ym_index = 0;
    yn_index = 0;
    ym       = 0;
    yn       = 1;  
    choose_qf((b-a) * eps);
    c_i.push(*this);    
  }

  static interval **peano;  
  static interval *peano_w;
  static int *peano_n;                             
  static const int m_min, m_max, m_stride, m_dim;

  class no_screennumber;

protected:

  static int m2m_index(int k){ return (k - m_min) /  m_stride; }  
  static int m_index2m(int k){ return k * m_stride + m_min; }
  static void init_data();
  static void free_data();

  void set_xcauchy(interval, int, int);
  void set_xcauchy(int, int, interval);

  void xsubdivide(real&, real&);
  void add_approxsum();
  void choose_qf(const real&);
  void subdivide(); 
  void print_interval(ostream& os){ 
          os  << " intervaltype: xcauchy " << *this << endl;}

 friend class cauchyxriemann_integral;
};



// ----------------------------------------------------------------------------
// --- class cauchyxriemann ---------------------------------------------------
// ----------------------------------------------------------------------------


class cauchyxriemann: public xcauchy, public xriemann{

 public:

 cauchyxriemann(real a, real b, real c, real d,  int k1, int l1, int k2, 
                int l2, cauchyxriemann_integral& c_i, bool t = false):
                xriemann(a, b, c, d, k1, l1, c_i, t),
                xcauchy(a, b, c, d, k1, l1, c_i, t),
                base_interval(a, b, c, d, k1, l1, c_i, t){

    xm       = xcauchy::m_index2m(xm_index);
    xn       = xriemann::n_index2n(xn_index);
    ym_index = k2;
    yn_index = l2;
    ym       = xriemann::m_index2m(ym_index);
    yn       = xriemann::n_index2n(yn_index);
    xdivide  = true;
    ydivide  = true;
  }
  cauchyxriemann(real a, real b, real c, real d,  int k1, int l1, int k2, 
        int l2, cauchyxriemann_integral& c_i, bool xdiv, bool ydiv, 
        real eps, bool t = false);


 protected:

  void add_approxsum();

  void choose_xcf(const real&);
  void choose_ycf(const real&);
  void choose_ycfeven(const real&);
  void choose_ycfodd(const real&);  

  void subdivide(); 
  //  void xsubdivide(real& left, real& right); 

  void set_xcaxri(int, int, interval);
  void set_xcaxri(interval, int, int); 
  void set_ycaxri(int, int, interval);
  void set_ycaxri(interval, int, int);
  void set_ycaxri(int, int, interval, interval);
  void set_ycaxri(interval, interval, int, int);
 
  void print_interval(ostream& os){ 
      if (transposed) os  << " intervaltype: riemann x cauchy ";
      else os  << " intervaltype: cauchy x riemann ";
      os << *this << endl;
  }

private:
  bool xdivide, ydivide;
  friend class cauchyxriemann_integral;
};



// ----------------------------------------------------------------------------
// --- class cauchy -----------------------------------------------------------
// ----------------------------------------------------------------------------


class cauchy:  public xcauchy{

 public:

  cauchy(real a, real b, real c, real d,  int k1, int l1, int k2, int l2, 
             cauchy_integral& c_i): base_interval(a, b, c, d, k1, l1, c_i),
                                    xcauchy(a, b, c, d, k1, l1, c_i)
 {
    xm       = m_index2m(xm_index);
    xn       = xriemann::n_index2n(xn_index);
    ym_index = k2;
    yn_index = l2;
    ym       = m_index2m(ym_index);
    yn       = xriemann::n_index2n(yn_index);    
  }

 protected:


  void add_approxsum();
  bool choose_cf(const real&);
  void subdivide(); 
 
  void ysubdivide(real&, real&);
  void remainders(interval&, interval&, interval&, interval&, interval&);
  real  choose_xcf(const real&);
  real  choose_ycfodd(const real&);
  real  choose_ycfeven(const real&);

  void set_tk(interval&, interval&, interval&, interval&, interval&);
  real set_xcauchy(interval, int, int);
  real set_xcauchy(int, int, interval);
  real set_ycauchy(interval, interval, int, int);
  real set_ycauchy(int, int, interval, interval = interval(0,0));


  void print_interval(ostream& os){ 
         os  << " intervaltype: cauchy " << *this << endl;}

   friend class cauchy_integral;
   friend class cauchyxriemann;
};



// ----------------------------------------------------------------------------
// --- class riemann ----------------------------------------------------------
// ----------------------------------------------------------------------------


inline riemann::riemann(real a, real b, real c, real d, int k1, int l1, 
     int k2, int l2, integral& r_i, bool xdiv, bool ydiv, real eps)
               : base_interval(a, b, c, d, k1, l1, r_i), 
                 xriemann(a , b, c, d, k1, l1, r_i)
{
  ym_index = k2;
  yn_index = l2;

  ym = m_index2m(ym_index);
  yn = n_index2n(yn_index);
    
  xdivide = xdiv;
  ydivide = ydiv;
  choose_xcf(eps);
  choose_ycf(eps);
  ri.push(*this);
}




// ----------------------------------------------------------------------------
// --- class xcauchy  ---------------------------------------------------------
// ----------------------------------------------------------------------------
// error classes:

class xcauchy::no_screennumber{
   public:
    no_screennumber(real a, real b, real c = 0.0, real d = 0.0){ 
       xlow = a; xup = b;
       ylow = c; yup = d;
    }
   private: 
     real xlow, xup, ylow, yup;
};




// ----------------------------------------------------------------------------
// --- class xcauchyxriemann --------------------------------------------------
// ----------------------------------------------------------------------------


inline cauchyxriemann:: cauchyxriemann(real a, real b, real c, real d,  
                 int k1, int l1, int k2, int l2, cauchyxriemann_integral& c_i,
				       bool xdiv, bool ydiv, real eps, bool t):
                xriemann(a, b, c, d, k1, l1, c_i, t),
                xcauchy(a, b, c, d, k1, l1, c_i, t),
                base_interval(a, b, c, d, k1, l1, c_i, t)
{
    xm       = xcauchy::m_index2m(xm_index);
    xn       = n_index2n(xn_index);
    ym_index = k2;
    yn_index = l2;
    ym       = xriemann::m_index2m(ym_index);
    yn       = n_index2n(yn_index);
    xdivide  = xdiv;
    ydivide  = ydiv;
    choose_xcf(eps);
    if (xn % 2) choose_ycfodd(eps);
    else choose_ycfeven(eps);
    c_i.push(*this);    
}


// ----------------------------------------------------------------------------
// --- class integral ---------------------------------------------------------
// ----------------------------------------------------------------------------
// inline memberfunctions:
// heap administration: push, pop, print_heap

inline base_interval* integral::pop(){
    pop_heap(heap, heap + heapsize--);
    accumulate(Rinf, -1.0, heap[heapsize].pointer->r_inf);
    accumulate(Rsup, -1.0, heap[heapsize].pointer->r_sup);
    return heap[heapsize].pointer;
}

inline void integral::push(base_interval& ri){
    heap[heapsize] = *new pbase_interval(ri);
    push_heap(heap, heap + ++heapsize); 
    accumulate(Rinf, 1.0, ri.r_inf);
    accumulate(Rsup, 1.0, ri.r_sup);
}

inline void integral::print_heap(){
    copy(heap, heap + heapsize, ostream_iterator<pbase_interval>(cout, "  "));
}

extern operand x, y;
