//============================================================================
//
//                              Program/Module
//                                   from
//                C++ package for verified numerical integration
//                  of singular and strong singular functions
//                     Copyright (c) 2000 Stefan Wedner
//
// This program/module is free software for non-commercial use. 
//
//
//============================================================================
//----------------------------------------------------------------------------
//
// File:    node.C
// Purpose: computing the nodes and weights of gaussian quadrature formulae
//          (with at least two nodes) using the recurrence relation
//     
//============================================================================


#include<string>
#include"node.h"

// ---------------------------------------------------------------------------
// static variables of the class
// ---------------------------------------------------------------------------

outformat symm::oflag1 = Hexa;  // default: print datas as hex!  
outformat symm::oflag2 = Nodes; // default: print nodes! 



// ---------------------------------------------------------------------------
// warnings
// ---------------------------------------------------------------------------

const string symm_warn[] = {
    "verify: No root in the interval! Increasing precision for bisection!",
    "unexpected case ",
    "verify: There is more than one root! Increasing number of iterations!",
    "verify: Verify failed! Increasing precison for calculating the coefficients!",
    "newton: Division by zero! Increasing number of iterations!",
    "weight: Division by zero! Improving enclosure of the nodes!",
    "invert: Division by zero! Increasing precision_coeff!",
    "qr    : Division by zero! Increasing precision_coeff!",
    "invert: Division by zero! Increasing precision_ival!",
    "qr    : Division by zero! Increasing precision_ival!",
};



// ---------------------------------------------------------------------------
// error messages
// ---------------------------------------------------------------------------

const string symm_error[] = {
  "verify: No root in the interval! Check startintveral and precision for bisection!",
  "Unexpected error!",
  "verify: There is more than one root! Increase number of iterations!",
  "verify: Verify failed! Increase precison for calculating the coefficients!",
  "newton: Division by zero! Increase number of  iterations or use extented inewton!",
  "weight: Division by zero! Improve enclosure of the nodes!",
  "invert: Division by zero! Increase precision_coeff and/or precision_ival! ",
  "qr    : Division by zero! Increase precision_coeff and/or precision_ival! ", 
  "weight: Your selfdefined class is incomplete! Define void xxx::h(int n) ! "         
};



// ---------------------------------------------------------------------------
// auxiliary functions
// ---------------------------------------------------------------------------

int Actual_Stagprec(l_interval& a){
 //  determine max stagprec needed for storing l_interval a
 //  0 < a

  real d = _real(diam(a)), sup = _real(Sup(a));
  int precision;

  if (d <= MinReal) return StagPrec(a);
  precision  = _double( ln(sup/d) / ln(10) ) / 15;
  return ++precision;
}



// l_rmatrix mid(l_imatrix A){
//   return Inf(A)*0.5+Sup(A)*0.5;
// }



// l_rvector mid(l_ivector b){
//   return Inf(b)*0.5+Sup(b)*0.5;
// }



void QR(l_rmatrix& A){
 // QR decomposition of a 2x2 matrix A
 // result is stored in A

  int dim = Ub(A, 1), i ,j, k;
  real r,s;
  l_rvector b(dim), help(dim);
  l_rmatrix Q(dim,dim);
  l_real norm = sqrt(A[Col(1)] * A[Col(1)]); 

  if (norm  == 0 ) throw qr_error();
  A[1][1] /= norm;
  A[2][1] /= norm;
  A[1][2] = - A[2][1];
  A[2][2] =   A[1][1];
}



void sort(l_rmatrix& B, l_ivector& z){
 // sort the columns, to start orthonormalization 
 // process with longest edge

  int dim = Ub(B, 1);
  l_rvector help(dim);
  if ( B[Col(1)] * B[Col(1)] * sqr( diam(z[1]) ) 
       < 
       B[Col(2)] * B[Col(2)] * sqr( diam(z[2]) ) )
  {
    help      = B[Col(1)];
    B[Col(1)] = B[Col(2)];
    B[Col(2)] = help;
  }
}



void invert(l_rmatrix& B, l_imatrix& Binv){
 // invert matrix B 

  l_interval det;
  det = B[1][1] * _l_interval(B[2][2]) - B[1][2] * _l_interval(B[2][1]);

  if (0 < det) throw invert_error();
  Binv[1][1] =   B[2][2] / det;
  Binv[1][2] = - B[1][2] / det;
  Binv[2][1] = - B[2][1] / det;
  Binv[2][2] =   B[1][1] / det;      
}



void multsub(l_imatrix& A, l_rvector& y_old, l_ivector& y_new){
 // compute enclosure of  A * y_old - y_new
 // result in y_new

  idotprecision idot;
 
  for(int i=1; i<= Ub(y_old); i++){
    idot = 0;
    accumulate(idot, y_old, A[Row(i)]);
    accumulate(idot, 1,-y_new[i]);
    y_new[i] =  idot;
  }
}


void multsub(l_imatrix& A, l_rvector& y_old, l_ivector& d, l_ivector& y_new){
 // compute enclosure of  d + A * y_old - y_new
 // result in y_new

 idotprecision idot;
 
 for(int i=1; i<= Ub(y_old); i++){
   idot = 0;
   accumulate(idot, y_old, A[Row(i)]);
   accumulate(idot, 1,-y_new[i]);
   accumulate(idot, 1, d[i]); 
   y_new[i] =  idot;
 }
}


// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// member functions symm
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------


// ---------------------------------------------------------------------------
// constructor and destructor
// --------------------------------------------------------------------------

symm::symm(int n){
  grad    = n;
  r_alpha = new l_real[n+1];
  r_gamma = new l_real[n+1];
  i_alpha = new l_interval[n+1]; 
  i_gamma = new l_interval[n+1];
  nodes   = new l_interval[n+1];
  weights = new l_interval[n+1];
}



symm::~symm(){
  delete[] r_alpha;
  delete[] r_gamma;
  delete[] i_alpha;
  delete[] i_gamma;
  delete[] nodes;
  delete[] weights;
}



// ---------------------------------------------------------------------------
// member functions for computing the recurrence equation (symm)
// ---------------------------------------------------------------------------

int symm::sturm(l_real  x){
  int sign_old = 1, sign_new = 1, sign_changes = 1;
  l_real p_old, p_new, p_temp, *alpha = r_alpha, *gamma = r_gamma;
 
  p_old =   *alpha;
  p_new =   - (*++alpha) * x; ++gamma;
  if (p_old < 0.0) sign_old = -1; 
  if (p_new < 0.0) sign_new = -1;
  else if  (p_new == 0.0) sign_new = sign_old;

  if (sign_old * sign_new < 0){ sign_changes++; sign_old = sign_new;}  

  for(int j=2 ;j<=grad;j++)
  { 
    p_temp=p_new;
    p_new= - (*++alpha) * x * p_new - (*++gamma) *  p_old;
    p_old=p_temp;

    if (p_new < 0) sign_new = -1;
    else if (p_new > 0) sign_new =  1;
    if (sign_old * sign_new < 0){ sign_changes++; sign_old = sign_new;}
  }
  return sign_changes;
}



int symm::sturm(l_interval  x){
  int sign_old = 0, sign_new = 0, sign_changes = 1;
  l_interval p_old, p_new, p_temp, *alpha = i_alpha, *gamma = i_gamma;
 
  p_old =   *alpha;
  p_new =   -  (*++alpha) * x ; ++gamma;

  sign_old = sign(p_old); // sign_old != 0
  if (sign_old == 2) return 0;
  sign_new = sign(p_new);
  if(sign_new ==  2) return 0;
  if (sign_old * sign_new < 0){sign_changes++; sign_old = sign_new;}  

  for(int j=2 ;j<=grad;j++)
  { 
    p_temp=p_new;
    p_new= - (*++alpha) * x * p_new - (*++gamma) *  p_old;
    p_old=p_temp;
    sign_new = sign(p_new);
    if(sign_new ==  2) return 0;
    if (sign_old * sign_new < 0){sign_changes++; sign_old = sign_new;}  
  }
  return sign_changes;
}



int symm::sturm(l_real x, l_interval& p_n){
 // computes p_n(x) (verified)
  l_interval p_old, p_temp, *alpha = i_alpha, *gamma = i_gamma;
 
  p_old =   *alpha;
  p_n =   (*++alpha) * x ; ++gamma;

  for(int j=2 ;j<=grad;j++)
  { 
    p_temp=p_n;
    p_n=  (*++alpha) * x * p_n - (*++gamma) *  p_old;
    p_old=p_temp;
  }
  return 0;
}



int symm::sturm(l_interval x, l_interval& p_n,l_interval& q_n ){
 // computes p_n(mid(x)) and p'_n(x)  (verified)
  l_interval p_old, pq_temp, q_old,
             *alpha = i_alpha,
             *gamma = i_gamma;
 
  p_old  =  *alpha;
  p_n    =  (*++alpha) * x ; ++gamma;
  q_old  =  l_interval(0,0);
  q_n    =  *alpha;

  for(int j=2 ;j<=grad;j++)
  { 
    pq_temp =  q_n;
    q_n     =  (*++alpha) * x * q_n + *alpha * p_n - (*++gamma) *  q_old;
    q_old   =  pq_temp;
    pq_temp =  p_n;
    p_n     =  (*alpha) * x * p_n - (*gamma) *  p_old;
    p_old   =  pq_temp;
  }
  sturm(mid(x), p_n);
  return 0;
}



int symm::sturmqr(l_real x){
 // computes changes of sign (verified) using QR decomposition
  int sign_old, sign_new, sign_changes = 1;
  l_interval *alpha = i_alpha, *gamma = i_gamma;   
  l_rvector y_old(2), y_new(2);
  l_ivector z_old(2), z_new(2), d_top(2);
  l_rmatrix B_old(2, 2), B_new(2, 2);
  l_imatrix A(2, 2), B_inv(2, 2);
   
  y_new[1] = mid(*alpha); 
  z_new[1] = y_new[1] - *alpha;     
  sign_old = sign(*alpha);
  if (sign_old == 2) return 0;
  y_new[2] = -mid(*++alpha) * x ; ++gamma;
  z_new[2] = y_new[2] + *alpha * x;
  sign_new = sign(- (*alpha) * x);
  if(sign_new ==  2) return 0;
  if (sign_old * sign_new < 0){sign_changes++; sign_old = sign_new;}  
  B_new[1][1] = 1;
  B_new[1][2] = 0;
  B_new[2][1] = 0;
  B_new[2][2] = 1;
  
  A[1][1] = _l_interval(0,0);
  A[1][2] = _l_interval(1,1);

  for(int i=2; i <= grad; i++){
    A[2][1] = -*++gamma;
    A[2][2] = -x * (*++alpha);  
    B_old = B_new;
    y_old = y_new;
    z_old = z_new;
    y_new = mid(A) * y_old;
    d_top = y_new;
    multsub(A, y_old, d_top);
    B_new = mid(A*B_old);
    sort(B_new, z_old);
    QR(B_new);
    invert(B_new, B_inv);
    z_new = B_inv * A * B_old * z_old + B_inv * d_top;
    z_old = y_new + B_new * z_new;
    sign_new = sign(z_old[2]);
    if(sign_new ==  2) return 0;
    if (sign_old * sign_new < 0){sign_changes++; sign_old = sign_new;}  
  }
  return sign_changes;
}



int symm::sturmqr(l_real x, l_interval& p_n){
  // computes p_n(x)(verified)  using QR decomposition
  l_interval *alpha = i_alpha, *gamma = i_gamma;   
  l_rvector y_old(2), y_new(2);
  l_ivector z_old(2), z_new(2), d_top(2);
  l_rmatrix B_old(2, 2), B_new(2, 2);
  l_imatrix A(2, 2), B_inv(2, 2);
   
  y_new[1] = mid(*alpha); 
  z_new[1] = y_new[1] - *alpha;     
  y_new[2] = mid(*++alpha) * x ; ++gamma;
  z_new[2] = y_new[2] - *alpha * x;
  B_new[1][1] = 1;
  B_new[1][2] = 0;
  B_new[2][1] = 0;
  B_new[2][2] = 1;
  
  A[1][1] = _l_interval(0,0);
  A[1][2] = _l_interval(1,1);

  for(int i=2; i <= grad; i++){
    A[2][1] = -*++gamma;
    A[2][2] =  x * (*++alpha);  
    B_old = B_new;
    y_old = y_new;
    z_old = z_new;
    y_new = mid(A) * y_old;
    d_top = y_new;
    multsub(A, y_old, d_top);
    B_new = mid(A * B_old);
    sort(B_new, z_old);
    QR(B_new);
    invert(B_new, B_inv);
    z_new = B_inv * A * B_old * z_old + B_inv * d_top;
  }
  z_old = y_new + B_new * z_new;
  p_n = z_old[2]; 
  return 0;
}



int symm::sturmqr(l_interval x, l_interval& q_n, l_interval& p_n){
 // computes p_n'(x) (for INewton) using QR decomposition
 
  l_interval l_p_n, alpha_i;
  l_interval *alpha = i_alpha, *gamma = i_gamma;  
  l_rvector y_old(2), y_new(2), y_deriv(2);
  l_ivector z_old(2), z_new(2), z_deriv(2), d_top(2), d_deriv(2);
  l_rmatrix B_old(2, 2), B_new(2, 2), B_deriv(2, 2);
  l_imatrix A(2, 2), B_inv(2, 2);
   
 // y_new[1], y_new[2] abhaengig von Gewichtsfunktion
  y_new[1] = mid(*alpha);
  z_new[1] = y_new[1] - *alpha;  
  y_new[2] = mid(*++alpha *x); ++gamma;
  z_new[2] = y_new[2] -  *alpha * x;
  B_new[1][1] = 1;
  B_new[1][2] = 0;
  B_new[2][1] = 0;
  B_new[2][2] = 1;
  
  y_deriv[1] = 0; 
  y_deriv[2] = mid(*alpha);
  B_deriv = B_new;
  z_deriv[1] = _l_interval(0,0);
  z_deriv[2] = y_deriv[2] - *alpha; 
  d_deriv[1] = 0;

  l_p_n   = y_new[2] + z_new[2];

  A[1][1] = _l_interval(0,0);
  A[1][2] = _l_interval(1,1);

  for(int i=2; i <= grad; i++){
    A[2][1] = -*++gamma;
    A[2][2] = x * *++alpha;  

   // compute p'_i(x)!
    B_old = B_deriv;
    y_old = y_deriv;
    z_old = z_deriv;
    d_deriv[2] = *alpha *  l_p_n;
    y_deriv = mid(A) * y_old + mid(d_deriv);
    d_top   = y_deriv;
    multsub(A, y_old, d_deriv, d_top);
    B_deriv = mid(A * B_old);
    sort(B_deriv, z_old);
    QR(B_deriv);
    invert(B_deriv, B_inv);
    z_deriv = B_inv * A * B_old * z_old + B_inv * d_top;
    p_n = l_p_n; // store p_{i-1} in p_n

    // compute p_i(x)
    B_old = B_new;
    y_old = y_new;
    z_old = z_new;
    
    y_new = mid(A) * y_old;
    d_top = y_new;
    multsub(A, y_old, d_top);
    B_new = mid(A*B_old);
    sort(B_new, z_old);
    QR(B_new);
    invert(B_new, B_inv);
    z_new = B_inv * A * B_old * z_old + B_inv * d_top;
    z_old = y_new + B_new * z_new;
    l_p_n = z_old[2];    
  }
   
  z_deriv = y_deriv + B_deriv * z_deriv;
  q_n = z_deriv[2];   
  return 0;
}



// ---------------------------------------------------------------------------
// member functions bisect, verify and newton (symm)
// ---------------------------------------------------------------------------

void symm::bisect(l_interval& start, int k){
 // Compute approximation interval by bisection
  stagprec = precision_real;
  l_real mitte=mid(start);
  int soc;

  for(int i=1; i<=iteration; i++){
    if(soc = sturm(mitte)){ 
      if(soc > k)    SetSup(start, mitte);
      else           SetInf(start, mitte);
      mitte = mid(start);
    }
  }  

  // cout << " bisect  " << stagprec << endl; 

}



int symm::verify(l_interval x, int k, int precision){
 // node x_k in [x.inf,x.sup) ?
 //  Exceptions:
 //  0:  x_k is NOT in  [x.inf,x.sup)
 //  2:  there are more nodes in [x.inf,x.sup)
 //  3:  uncertain (increase the precision?)
 //  NO exception -->  x_k is the unique node in [x.inf,x.sup)

  int sign_change = 0, //   changes of sign not determined 
      multiple = 0;
  for(stagprec = precision; stagprec <= precision_coef ;  stagprec++){
    if (sign_change = sturmqr(Inf(x))){
      if (sign_change <= k) break;
      else throw verify_error(0);
    }
  }
  if (!sign_change) throw verify_error(3);
  if (sign_change < k) multiple =1;
   // use same precision (or higher) to verify sup
  for(; stagprec <= precision_coef ; stagprec++)
   if (sign_change = sturmqr(Sup(x))){
     if (sign_change > k) break;
     else throw verify_error(0);  
   }
  if (!sign_change) throw verify_error(3) ;
  if (sign_change > k+1 || multiple) throw verify_error(2);
  return stagprec;   
}



int  symm::newton(l_interval& node){
 // Improve enclosure  Newton
 // 0        :   0 is in  p'_n
 // returns number of iteration 
   
  int i=0, j;
  l_interval x_old, x_new, p_n, q_n;

  x_new = node;
  stagprec = precision_ival; 
  do{        
      x_old = x_new;	  
      sturmqr(x_old, q_n, p_n); 
      sturmqr(mid(x_old), p_n);
      if(0 <= q_n) throw newton_error();
      x_new =(mid(x_old) - p_n / q_n) & x_old;
      i++;       
    }
    while( x_new != x_old && i < max_newton); 
    node = x_new;


    // cout << " newton  " << stagprec; 



    return i;
}



// ----------------------------------------------------------------------------
// member functions for computing the scalarproducts (p_i, p_i) (symm)
// ----------------------------------------------------------------------------

void symm::calc_scalps(bool use_f){
  if (orthonormal) return;
  if( !use_f ){
    l_interval h_j = *hscalp * i_alpha[1] / i_alpha[0] ;
    for(int j = 1; j < grad; j++){
      h_j *= i_gamma[j+1];
      hscalp[j] = h_j / i_alpha[j+1];
    }  
  }
  else      // use function for h_i
    for(int k = 0; k < grad; k++) hscalp[k] = h(k);
}



l_interval symm::h(int i){
 // computes (p_i, p_i)
 // used if orthonormal = false
 //         *hscalp contains (p0, p0)
 // otherwise h is overwritten

  if (*hscalp == _l_interval(0,0)) throw incomplete_error(1);
  if(i==0) return *hscalp;
  l_interval h_j = *hscalp * i_alpha[1] / i_alpha[0] ;
  for(int j = 1; j <= i; j++)
    h_j *= i_gamma[j+1];
  return h_j / i_alpha[i];
}



// ----------------------------------------------------------------------------
// member functions for computing the weights (symm)
// ----------------------------------------------------------------------------

l_interval symm::weight1(l_interval& x){
 // Compute weight for node x using:
 // w = alpha_n * (p_n-1, p_n-1) / ( p'_n(x) * p_n-1(x) )
  l_interval p_n, q_n, weight;
  stagprec = precision_ival;
  sturmqr(x, q_n, p_n);
  if (0 <= p_n || 0 <= q_n) throw weight_error();
  weight =  1 / ( p_n * q_n ); 
  if (orthonormal) return i_alpha[grad] * weight;
  return hscalp[grad -1] * i_alpha[grad] * weight;
}



l_interval symm::weight2(l_interval& x){
 // Compute weight for node x using 
 // w  = 1 / ( sqr( p_0(x) ) / h_0 + .... + sqr( p_{n-1} ) / h_{n-1} )
 // with h_i = (p_i, p_i)

  stagprec = precision_ival;
  l_interval *alpha = i_alpha, *gamma = i_gamma,
             result, p_i, h_i;
  l_rvector y_old(2), y_new(2);
  l_ivector z_old(2), z_new(2), d_top(2);
  l_rmatrix B_old(2, 2), B_new(2, 2);
  l_imatrix A(2, 2), B_inv(2, 2);
      
   y_new[1] = mid(*alpha); 
   z_new[1] = y_new[1] - *alpha;  
   result   = *alpha;   
   y_new[2] = mid(*++alpha * x) ; ++gamma;
   z_new[2] = y_new[2] - *alpha * x;
   B_new[1][1] = 1;
   B_new[1][2] = 0;
   B_new[2][1] = 0;
   B_new[2][2] = 1;
  
   A[1][1] = _l_interval(0,0);
   A[1][2] = _l_interval(1,1);
   p_i     = *alpha * x;

  if (orthonormal) result = sqr(i_alpha[0]) + sqr(p_i);
  else result = sqr(i_alpha[0]) / *hscalp + sqr(p_i) / hscalp[1];

  for(int i=2; i < grad; i++){ 
    A[2][1] = -*++gamma;
    A[2][2] = x * (*++alpha);  
    B_old = B_new;
    y_old = y_new;
    z_old = z_new;
    y_new = mid(A) * y_old;
    d_top = y_new;
    multsub(A, y_old, d_top);
    B_new = mid(A*B_old);
    sort(B_new, z_old);
    QR(B_new);
    invert(B_new, B_inv);
    z_new = B_inv * A * B_old * z_old + B_inv * d_top;
    z_old = y_new + B_new * z_new;
    p_i = z_old[2];
    if (orthonormal) result += sqr(p_i);
    else result += sqr(p_i) / hscalp[i];
  }
  if (0 <= result) throw weight_error();
  return 1.0 / result; 
}



// ----------------------------------------------------------------------------
// member function calc_coefficients, verify  (public versions)
// ----------------------------------------------------------------------------

void symm::calc_coefficients(int precision){
 precision_coef = stagprec = precision;
 calc_coefficients();
}



int symm::verify(l_interval x, int k, int precision, int prec_max){
  if( prec_max < precision) prec_max = precision > 1 ? precision: 2; 
  precision_coef = prec_max; 
  verify(x, k, precision);
}



// ----------------------------------------------------------------------------
// member functions calc_node*
// ----------------------------------------------------------------------------

int symm::calc_node(l_interval& enclosure, int i){
 // returns precision needed for verifikation

  int precision  = precision_real, result;
                                         
  precision_ival = calc_coefficients();
  bisect(enclosure, i);
  precision = precision_real;
  result =  verify(enclosure, i, precision);
  newton(enclosure);
  return result;
}

int symm::calc_node_and_weight(l_interval& enclosure, int i, 
                               int prec_coef, int prec_real, 
                               int iter, ostream& os){
 // public version 
  int precision;
  l_interval Weight;
  if(prec_real < 1 || prec_coef < 1){
    os << " Invalid input, using default precisions! " << endl;
    prec_coef = 2; 
    prec_real = 2; 
  }

  precision_coef = stagprec = prec_coef; 
  precision_real = prec_real; 

  iteration      = iter;
  try{
    precision = calc_node(enclosure, i);
    os << " Node  : " << endl << enclosure << " " << diam(enclosure) << endl;
    //    Weight = weight1(enclosure);
    //  os << " Weight: " << endl << Weight << " " << diam(Weight) << endl;
    Weight = weight2(enclosure);
    os << " Weight: " << endl << Weight << " " << diam(Weight) << endl;
  }
  catch (newton_error){
    os << " Error " << symm_error[4] << endl;
  }
  catch (weight_error){
    os << " Error " << symm_error[5] << endl;
  }
  catch (verify_error v){
    os << " Error " << symm_error[v.i] << endl;
  }
  catch (qr_error){
    os << " Error " << symm_error[7] << endl;
  }
  catch (invert_error){
    os << " Error " << symm_error[6] << endl;
  }     
  catch (incomplete_error m){
     os << " Error "   << symm_error[7+m.i] << endl;
  } 
  return precision;
}



void symm::calc_nodes_and_weights(const l_interval& start, int i, int j, 
                                                    ostream& os, istream& in){
 // public  interactive version
  l_interval enclosure, Weight;
  
  if(!j) j = i; 
  while(i <= j){
    // input 0 for step over
  try{
    enclosure = start;
    os << " input precision for computing coefficients: "; in >> precision_coef;
    if( precision_coef){
      stagprec = precision_coef; 
    }
    os << " precision of coefficients is: " << calc_coefficients()  << endl;
    os << " input precision for real bisection: "; in >> precision_real;
    os << " input number of iterations (real bisection): "; in >> iteration;
    if( iteration && precision_real){
    os << " ------ Computing node and weight ------" << endl; 
    bisect(enclosure, i);
      os << " real bisection: " << endl << enclosure << endl;
      os << " node verified with precision " <<  verify(enclosure, i, precision_real) << endl;
    }
    os << " input precision for inewton  "; in >> precision_ival;
    if (precision_ival){
      os << " after " << newton(enclosure) << " Newton steps " << endl;
      os << enclosure << " diam " << diam(enclosure) << endl;
    }
    os << " input precision for computing weight: "; in >> precision_ival; 
    if (precision_ival){
      // Weight = weight1(enclosure);
      // os << " weight1: " << endl << Weight << " " << diam(Weight) << endl;
      Weight = weight2(enclosure);
      os << " weight: " << endl << Weight << " " << diam(Weight) << endl;
    }
    i++;
  }
  catch (newton_error){
    os << " Error " << symm_error[4] << endl;
  }
  catch (weight_error){
    os << " Error " << symm_error[5] << endl;
  }
  catch (verify_error v){
    os << " Error " << symm_error[v.i] << endl;
  }
  catch (qr_error){
    os << " Error " << symm_error[7] << endl;
  }
  catch (invert_error){
    os << " Error " << symm_error[6] << endl;
  }     
  catch (incomplete_error m){
     os << " Error "   << symm_error[7+m.i] << endl;
  }
 
  }
}



int symm::calc_node_and_weight(int iter, int prec_coef, int prec_real,
                               l_interval& encl, int i, ostream& os)
{
 // private version
 // compute nodes i and weight i
 // return 1 ferror 
 // return 0 success
  int precision = precision_real;
  l_interval help, enclosure;

  while(1){
    try{
      enclosure = encl;
      bisect(enclosure, i);   
      os << " ------ Computing node and weight " << i << " ------" << endl; 
      os << " real bisection: " << endl << enclosure << endl;
      verify(enclosure, i, precision);  
      os << " node verified with precision " << precision << endl; 
      os << enclosure << " diam " << diam(enclosure) << endl;
      os << " after " << newton(enclosure) << " Newton steps " << endl;
      os << enclosure << " diam " << diam(enclosure) << endl;
      encl = enclosure;  
      //help = weight1(enclosure);
      //os << " weight1 " << endl << help;
      //os << " diam "    << diam(help) << endl;
      help = weight2(enclosure);
      os << " weight " << endl << help;
      os << " diam "    << diam(help) << endl << endl;
      nodes[i - 1]   = enclosure;
      weights[i - 1] = help;
      return 0;
    }
   // errorhandling
  catch(newton_error){ 
    if (iteration <= iter){
      os << " Warning " << symm_warn[4] << endl;
      iteration+=10;
    }
    else{
      os << " Error " << symm_error[4] << endl; 
      return 1;
    }
  }
  catch (weight_error){
    if(precision_ival < prec_coef){ precision_ival++;
      os << " Warning " << symm_warn[5] << endl; 
    }
    else if(precision_coef < prec_coef)
           { 
           stagprec       = ++precision_coef; 
           precision_ival = calc_coefficients();
           os << " Warning " << symm_warn[5] << endl; 
           }
         else{
            os << " Error " << symm_error[5] << endl;
            return 1;
	  }
  }
  catch (verify_error v){
    switch(v.i){
      case 0: 
            if (precision_real < prec_real){
               if (precision_ival <= precision_real)
	         if (precision_coef < prec_coef){
                   os << " Warning "   << symm_warn[v.i] << endl;
                   stagprec       = ++precision_coef;
                   precision_ival = calc_coefficients();
		 }
                 else{
                   os << " Error "   << symm_error[v.i] << endl;
                   return 1;
		 }
              os << " Warning " << symm_warn[v.i] << endl;
              precision_real++; break;  
	    }
            else{
              os << " Error "   << symm_error[v.i] << endl;
              return 1;   
	    }
 
      case 2:
            if (iteration < iter){ iteration+=iteration_step;
              os << " Warning " << symm_warn[v.i] << endl; break;
	    }
            else{  
              os << " Error "   << symm_error[v.i] << endl;
              return 1; 
            }
      case 3: 
            if (precision_coef < prec_coef){
             os << " Warning " << symm_warn[v.i] << endl;
             stagprec       = ++precision_coef;
             precision_ival = calc_coefficients();
	    }
            else{
              os << " Error "   << symm_error[v.i] << endl;
              return 1;
	    }
            break;
      default: os <<  symm_error[1] << endl;
      }
    }
    catch (qr_error){
      if (precision_ival < prec_coef) precision_ival++;
      else if(precision_coef < prec_coef){
        stagprec       = ++precision_coef;
        precision_ival = calc_coefficients();
      }
      else{
        os << " Error "   << symm_error[7] << endl;
        return 1;
      }
      os << " Warning " << symm_warn[7] << endl;     
    }
    catch (invert_error){
      if (precision_ival < prec_coef) precision_ival++;
      else if(precision_coef < prec_coef){
        stagprec = ++precision_coef;
        precision_ival = calc_coefficients();
      }
      else{
        os << " Error "   << symm_error[7] << endl;
        return 1;
      }
      os << " Warning " << symm_warn[7] << endl;   
    }
    catch (incomplete_error m){
      os << " Error "   << symm_error[7+m.i] << endl;
      return 1;
    }
  }
  throw fatal_error();
  return 0;
}



int symm::calc_nodes_and_weights(int prec_coef, int prec_real, 
                                 int iter, ostream& os)
{
 // init precisions 
  if(prec_real < 1 || prec_coef < 1){
    os << " Invalid input, using default precisions! " << endl;
    prec_coef = 2; 
    prec_real = 2; 
  }
  precision_coef = stagprec = prec_coef; 
  precision_real = prec_real; 
  precision_ival = calc_coefficients();

  // init count of iterations
 
   if(iter<=0){ 
     os << " Invalid input, using defaults! " << endl; 
     iteration = iter = 30;
   }
   else iteration = iter;

 // max values for error handling
  iter      +=  iteration_corr;
  prec_coef += precision_coef_corr;
  prec_real += precision_real_corr;

 // local variables
  l_interval enclosure(inf, sup), help;

  for(int i = 1; i <= n_max;){
  // compute node i !
    if ( calc_node_and_weight(iter, prec_coef, prec_real, enclosure, i, os) ) 
         return i;
    i++; 
    enclosure = _l_interval(Sup(enclosure), sup);  
  }  
  if (grad - 2 * n_max == 1){  
    enclosure = _interval(0,0);
    while(1){
      try{
      help=weight2(enclosure); 
      nodes[n_max]   = enclosure;
      weights[n_max] = help; 
      }
      catch (weight_error){
        if( precision_ival < prec_coef ){ 
          precision_ival++;
          os << " Warning " << symm_warn[5] << endl; 
	}
        else if(precision_coef < prec_coef){ 
          stagprec       = ++precision_coef; 
          precision_ival = calc_coefficients();
          os << " Warning " << symm_warn[5] << endl; 
	}
          else{
            os << " Error " << symm_error[5] << endl;
            return grad/2;
	  }
      }
      os << " ------ Computing node and weight " <<  grad/2 + 1 << " ------"
         << endl;
      os << " node nothing to verify " << endl;   
      os << " weight: " << endl << help;
      os << " diam "    << diam(help) << endl; 
      return grad;
    }
  }
  return grad;
}


 
// ---------------------------------------------------------------------------
// member functions (test functions) 
// --------------------------------------------------------------------------- 

void symm::test_coeff(ostream& os){
  l_real      *pr_alpha = r_alpha, *pr_gamma = r_gamma;
  l_interval  *pi_alpha = i_alpha, *pi_gamma = i_gamma;
  os << " Eingabe von stagprec ";
  cin >> stagprec;
  calc_coefficients();
  cout << " stagprec = " << stagprec << endl;
  for(int i=0; i<=grad; i++)
    os << " i_alpha[" << i << "] " << pi_alpha[i] << " diam " 
       << diam(*i_alpha) << endl   << " i_gamma[" << i 
       << "] " << pi_gamma[i] << " diam " << diam(*i_gamma) << endl
       << " r_alpha[" << i << "] " << pr_alpha[i] << endl
       << " r_gamma[" << i << "] " << pr_gamma[i] << endl << endl; 
}



void symm::test_sturm(ostream& os){
  real dr;
  interval pr;
  l_real x;
  l_interval p_n;
  int n, prec;
  os  << " Eingabe x : ";
  cin  >> x;
  os << " Eingabe prec ";
  cin  >> prec;
  stagprec = prec;
  n = grad;
  calc_coefficients();
  for(int i = 2; i <= n; i++){
  decrease_grad(i);
  sturm(x, p_n ); pr = diam(p_n);
  os << i << " & ";
  os << p_n << " & " << Scientific <<  SetPrecision(10,5)  << Inf(pr) << " & ";
  sturmqr(x, p_n);  pr = diam(p_n);
  os << p_n << " & " <<  SetPrecision(10,5)  << Inf(pr) << " & " << endl;
  }
}



l_interval symm::test_nodes_and_weights(int n, int precision){

  stagprec = precision;
  l_interval quadraturesum(0,0);
  if(n){
    for(int i=0; i < n_max; i++)
      quadraturesum+=weights[i] *  power(nodes[i],n);
    if ( n_max < grad)  for(int i=0; i < n_max; i++)
      quadraturesum+=weights[i] *  power(-nodes[i],n);
    if(grad - 2 * n_max == 1)  
       quadraturesum+=weights[grad/2] *  power(nodes[grad/2],n);
  }
  else{ 
    for(int i=0; i < n_max; i++)
      quadraturesum+= weights[i];
    if ( n_max < grad)  quadraturesum *= 2;
    if(grad - 2 * n_max == 1) quadraturesum+=weights[grad/2];
  }  
  return quadraturesum;
}



// ---------------------------------------------------------------------------
// friend functions
// ---------------------------------------------------------------------------

ostream& operator<<(ostream& os, const symm&  out){
  l_interval* pointer;
  interval help;
  int n;

  if( 2 * out.n_max < out.grad) n = out.n_max + 1;
  else                          n = out.n_max;

  if (out.oflag2 == Nodes) pointer = out.nodes;
  else pointer = out.weights;
  if (out.oflag1 == Hexa){
    for(int i=1 ; i <= n; i++){
      help = _interval(*pointer++); // remove underscore for new C-XSC      
      os << Hex << Inf(help) << endl << Sup(help) << Dec << endl;
    }
  }
  else if (out.oflag1 == Decimal){
    for(int i=1 ; i <= n ; i++){
      help = _interval(*pointer++); // remove underscore for new C-XSC      
      os  << Dec << help << endl  << diam(help) << endl;
    }
  }
  else{    // out.oflag1 Long
    for(int i=1 ; i <= n; i++){
      string tmp; // workaround avoid concatenation!!
      tmp <<  SetDotPrecision(50,45) << DotFormat << *pointer++;
      os << tmp << endl;
    }
  }
  return os;
}



ostream& operator<<(ostream& os, outformat out){
  if (out == Nodes || out == Weights) symm::oflag2 = out;
  else symm::oflag1 = out;
  return os;
}



// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// derived classes (symm)
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------



// ---------------------------------------------------------------------------
// class legendre
// ---------------------------------------------------------------------------

legendre::legendre(int n):symm(n){
 // init class data

  n_max       = n / 2;
  inf         = -1;
  sup         =  1;
  orthonormal = false;
  hscalp      = new l_interval[n];
}



legendre::~legendre(){
  delete[] hscalp;
}



int legendre::calc_coefficients(){
  
  l_interval l_itel;
  l_interval *pi_alpha = i_alpha, *pi_gamma = i_gamma; 
  l_real     *pr_alpha = r_alpha, *pr_gamma = r_gamma; 

  *pi_alpha = 1;
  *pr_alpha = 1;
  *pi_gamma = 0;
  *pr_gamma = 0; 
   for(int i = 1 ; i <= grad ; i++){
     l_itel      = i; 
     l_itel      = 1 / l_itel;
     *++pi_alpha = 2 - l_itel;
     *++pr_alpha = mid(*pi_alpha);
     *++pi_gamma = 1 - l_itel;
     *++pr_gamma = mid(*pi_gamma);
   }
  calc_scalps(true);
  return stagprec;
}



l_interval legendre::h(int i){
  l_interval l_i;
  l_i = i;
  return 2 / ( 2 * l_i + 1);
}



l_interval legendre::weight(l_interval& node){
 // Compute weight for node
 
  l_interval l_j, p_old(1,1), p_new = node, temp, temp2;
           
  temp2 = 0.5;
  for(int j=2;j<= grad;j++){
    l_j   = j;
    temp  = p_new; 
    p_new = (1-1/l_j) * (node * p_new - p_old) + node * p_new;
    p_old = temp;
    temp2  = (l_j - 0.5) * sqr(p_old) + temp2;
  }
  if (0 <= temp2) throw weight_error() ;
     return 1 / temp2;
}



// ---------------------------------------------------------------------------
// class coeff_symm
// ---------------------------------------------------------------------------



coeff_symm::coeff_symm(int n, real max, init_function alpha, 
		       init_function gamma, l_interval h0):symm(n)
{

  n_max          = n / 2;
  inf            = -max;
  sup            =  max;

  if (h0 == 0)
    orthonormal = true;
  else{
    orthonormal  = false;
    hscalp       = new l_interval[n];
    *hscalp      = h0;
  }
   
  init_alpha     = alpha;
  init_gamma     = gamma;
}



coeff_symm::~coeff_symm(){
  if(!orthonormal) delete[] hscalp;
}



int coeff_symm::calc_coefficients(){

 l_interval *pi_alpha = i_alpha, *pi_gamma = i_gamma; 
 l_real     *pr_alpha = r_alpha, *pr_gamma = r_gamma; 

  
  *pi_alpha = init_alpha(0);
  *pr_alpha = mid(*pi_alpha);
  *pi_gamma = 0;
  *pr_gamma = 0;  
  for(int i = 1 ; i <= grad ; i++){
     *++pi_alpha = init_alpha(i);
     *++pr_alpha = mid(*pi_alpha);
     *++pi_gamma = init_gamma(i);
     *++pr_gamma = mid(*pi_gamma);
  }
  calc_scalps(false);
  return stagprec;
}



// ---------------------------------------------------------------------------
// class gautschi_symm
// ---------------------------------------------------------------------------

gautschi_symm::gautschi_symm(int n, real max, init_function calc_moments)
 :symm(n){
  //  generic type with init functions for moments

  n_max          = n / 2;
  sup            = max;
  inf            = -max;
  orthonormal    = true;
  init_moment    = calc_moments;
}



int gautschi_symm::calc_coefficients(){
// computes coefficients from ordinary moments


  int n = 2* grad;
  idotprecision dot;
  l_interval*  c = new l_interval[n + 1];
  l_interval**  r = new l_interval*[grad + 1]; // (r)_ij = R transposed
  for(int i=0; i<=n;i++) c[i] = init_moment(i);
  for(int i=0; i<=grad;i++) r[i] = new l_interval[i + 1];

  r[0][0] = sqrt(c[0]); // j = 0    
  for(int j=1; j<=grad; j++){
    for(int i=0; i<j; i++){
      if ( (i+j) % 2 ) r[j][i] = 0;
      else{
      dot = c[i+j];
      for(int k=0; k<i; k++)
        accumulate(dot, -r[i][k], r[j][k]);
      r[j][i] = dot;
      r[j][i] /= r[i][i];
     }
    }
    dot = c[2*j];
    for(int k=0; k<j; k++)
      accumulate(dot, -r[j][k], r[j][k]);
    r[j][j] = dot;
    r[j][j] = sqrt(r[j][j]);
  }

  l_interval *pi_alpha = i_alpha, *pi_gamma = i_gamma; 
  l_real     *pr_alpha = r_alpha, *pr_gamma = r_gamma; 

  *pi_alpha = 1 / r[0][0];
  *pr_alpha = mid(*pi_alpha);
  *pi_gamma = 0; // not used
  *pr_gamma = 0; // not used

  *++pi_alpha = 1 / r[1][1] ; // p_1 = alpgha_1 * x 
  *++pr_alpha = mid(*pi_alpha);
  *++pi_gamma = 0; // not used;
  *++pr_gamma = 0; // not used;

  for(int i = 2 ; i <= grad ; i++){
    *++pi_alpha = r[i-1][i-1] / r[i][i];
    *++pr_alpha = mid(*pi_alpha);
    *++pi_gamma = sqr( r[i-1][i-1] ) / ( r[i][i] * r[i-2][i-2] );
    *++pr_gamma = mid(*pi_gamma);
   } 
   
   for(int i=0; i<=grad; i++) delete[] r[i];
   delete[] r;

   return Actual_Stagprec(*pi_alpha); 
}



// ---------------------------------------------------------------------------
// class sado_symm
// ---------------------------------------------------------------------------

sado_symm::sado_symm(int n, real max, init_function alpha, init_function gamma,
                     init_function calc_moments, bool calc_m):symm(n)
{
 // generic type for modified moments

  sup             =  max;
  inf             = -max;
  n_max           = n / 2;
  orthonormal     = false;
  ordinary        = calc_m;
  init_alpha      = alpha;
  init_gamma      = gamma;
  init_moment     = calc_moments;
  hscalp          = new l_interval[n];
  
}



sado_symm::~sado_symm(){
  delete[] hscalp;
}


 
void sado_symm::calc_mod(l_interval** s, l_interval* alpha_mod, 
                         l_interval* gamma_mod){
 //  computes the modified moments from the ordinary moments
 //  and the start values for the Sack/Donovan iteration

  //  stagprec++;  // !!!!!!!!!!! remove comments

  int i,j;
  l_interval** const nu = new l_interval*[2*grad];

  nu[0] = new l_interval[2 * grad];
  for(j=0; j < 2*grad; j++) nu[0][j] = init_moment(j);
  s[0][0] = 1;
  nu[1] = new l_interval[2 * grad - 1];
  for(j=0; j < 2*grad-1; j++) 
    nu[1][j] = alpha_mod[1] * nu[0][j+1]; 
  s[0][1] = nu[1][0] / nu[0][0];
  for(i=2; i < 2*grad; i++){
    nu[i] = new l_interval[2 * grad  - i];
    for(j = 0; j < 2 * grad - i; j++)
      nu[i][j] =  alpha_mod[i] * nu[i-1][j+1] - gamma_mod[i] * nu[i-2][j];
    s[0][i] = nu[i][0] / nu[0][0];  
  }
  for(i=0; i < 2*grad; i++) delete[]  nu[i];
  delete[] nu;

  //  stagprec --; remove comments
}



int sado_symm::calc_coefficients(){
 // Sack/Donovan iteration: 
 // computes the coefficients from the modified moments
 // and coefficients


  int i, j, n = grad;

  l_real     *pr_alpha  = r_alpha, *pr_gamma = r_gamma; 
  l_interval *pi_alpha  = i_alpha, *pi_gamma = i_gamma;
  l_interval rho;
  l_interval *alpha_mod = new l_interval[2*n], 
             *gamma_mod = new l_interval[2*n];
  l_interval** const s = new l_interval*[n];

 // memory allocation rows s[i]
  for(j=0; j<n; j++)
    s[j] = new l_interval[2 * grad  - j]; 
 
 // init coefficients alpha__mod, gamma_mod
  for(i=0; i<=2*n-1;i++){
    alpha_mod[i] = init_alpha(i);
    gamma_mod[i] = init_gamma(i);
  } 

 // store coefficients alpha_0, gamma_0, gamma_1
  *pi_alpha   = 1.0;
  *pr_alpha   = 1.0;
  *pi_gamma   = 0; 
  *pr_gamma   = 0;
  *++pi_gamma = 0; 
  *++pr_gamma = 0;

  if(ordinary){ 
    calc_mod(s, alpha_mod, gamma_mod);
    *hscalp     = init_moment(0);
  }
  else{  //  compute the modified moments using the init function! 
    *hscalp = init_moment(0);
    *hscalp /= alpha_mod[0];
    for(j=0; j <= 2 * n - 1; j++)
       s[0][j] = init_moment(j) / *hscalp;
  }

  rho   = ( s[0][2] + gamma_mod[2] ) / alpha_mod[2]; 
   // store coefficients alpha_1, gamma_2,
  *++pi_alpha = 1.0;
  *++pr_alpha = 1.0;
  *++pi_gamma = rho / alpha_mod[1]; 
  *++pr_gamma = mid(*pi_gamma);     
   // compute s[1][*]
  s[1][0] = ( s[0][1] / alpha_mod[1] ) / rho;
  for(j=1; j <= 2 * n - 2; j++)
    s[1][j] =  (s[0][j+1]  + gamma_mod[j+1] * s[0][j-1]) / alpha_mod[j+1] / rho;

  for(i=1 ; i <= n - 2; i++){
   // compute rho_i;
    rho   =  ( s[i][i+2] + gamma_mod[i+2] ) / alpha_mod[i+2] -
               s[i-1][i+1] / alpha_mod[i];  
   // store coefficients alpha_i+1, gamma_i+2,
    *++pi_alpha = 1.0;
    *++pr_alpha = 1.0;
    *++pi_gamma = rho / alpha_mod[i+1]; 
    *++pr_gamma = mid(*pi_gamma);
     
   // compute s[i+1][*]
    s[i+1][0] =( s[i][1]  / alpha_mod[1] -  s[i-1][0] / alpha_mod[i] ) / rho;
    for(j=1; j <= 2 * n - 2 - i; j++)
      s[i+1][j] =( (s[i][j+1]  + gamma_mod[j+1] * s[i][j-1]) / alpha_mod[j+1]-
                    s[i-1][j] / alpha_mod[i] ) / rho;
  }
   // store alpha_n 
  *++pi_alpha = 1.0;
  *++pr_alpha = 1.0;

   // free memory
  delete[] alpha_mod;
  delete[] gamma_mod;
  for(i=0; i<n; i++) delete[]  s[i];
  delete[] s;
  calc_scalps(false);
  return Actual_Stagprec(*pi_gamma); 
}



// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
// --------------------- class not_symm ----------------------------------------
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------



// -----------------------------------------------------------------------------
// --------------------- constructor, destructor -------------------------------
// -----------------------------------------------------------------------------



not_symm::not_symm(int n):symm(n){
  r_beta = new l_real[n+1];
  i_beta = new l_interval[n+1];
}



not_symm::~not_symm(){
  delete[] r_beta;
  delete[] i_beta;
}



// ---------------------------------------------------------------------------
// member functions for computing the recurrence equation (not_symm)
// ---------------------------------------------------------------------------

int not_symm::sturm(l_real  x){

  int sign_old = 1, sign_new = 1, sign_changes = 1;

  l_real p_old, p_new, p_temp, *alpha = r_alpha, 
         *beta = r_beta, *gamma = r_gamma;
 
  p_old =   *alpha;
  p_new =  (*++beta)  - (*++alpha) * x ; ++gamma;
  if (p_old < 0.0) sign_old = -1; 
  if (p_new < 0.0) sign_new = -1;
  else if  (p_new == 0.0) sign_new = sign_old;

  if (sign_old * sign_new < 0){ sign_changes++; sign_old = sign_new;}  

  for(int j=2 ;j<=grad;j++){ 
    p_temp=p_new;
    p_new= (*++beta - (*++alpha) * x) * p_new - (*++gamma) *  p_old;
    p_old=p_temp;

    if (p_new < 0) sign_new = -1;
    else if (p_new > 0) sign_new =  1;
    if (sign_old * sign_new < 0){ sign_changes++; sign_old = sign_new;}
  }
  return sign_changes;
}



int not_symm::sturm(l_interval  x){
  int sign_old = 0, sign_new = 0, sign_changes = 1;
  l_interval p_old, p_new, p_temp, *alpha = i_alpha, 
             *beta = i_beta, *gamma = i_gamma;
 
  p_old =   *alpha;
  p_new =   *++beta -  (*++alpha) * x ; ++gamma;

  sign_old = sign(p_old); // sign_old != 0
  if (sign_old == 2) return 0;
  sign_new = sign(p_new);
  if(sign_new ==  2) return 0;
  if (sign_old * sign_new < 0){sign_changes++; sign_old = sign_new;}  

  for(int j=2 ;j<=grad;j++){ 
    p_temp=p_new;
    p_new= (*++beta - (*++alpha) * x) * p_new - (*++gamma) *  p_old;
    p_old=p_temp;
    sign_new = sign(p_new);
    if(sign_new ==  2) return 0;
    if (sign_old * sign_new < 0){sign_changes++; sign_old = sign_new;}  
  }
  return sign_changes;
}



int not_symm::sturm(l_real x, l_interval& p_n){
 // computes p_n(x) (verified)
  l_interval p_old, p_temp, *alpha = i_alpha,
             *beta = i_beta, *gamma = i_gamma;
 
  p_old =   *alpha;
  p_n =   (*++alpha) * x - (*++beta); ++gamma;

  for(int j=2 ;j<=grad;j++)
  { 
    p_temp=p_n;
    p_n=  ( (*++alpha) * x + (*++beta) ) * p_n - (*++gamma) *  p_old;
    p_old=p_temp;
  }
  return 0;
}



int not_symm::sturm(l_interval x, l_interval& p_n,l_interval& q_n ){
 // computes p_n(mid(x)) and p'_n(x)  (verified)
  l_interval p_old, pq_temp, q_old, *alpha = i_alpha, 
             *beta = i_beta, *gamma = i_gamma;
 
  p_old  =  *alpha;
  p_n    =  (*++alpha) * x + (*++beta); ++gamma;
  q_old  =  l_interval(0,0);
  q_n    =  *alpha;

  for(int j=2 ;j<=grad;j++)
  { 
    pq_temp =  q_n;
    q_n     =  ( (*++alpha) * x + (*++beta) ) * q_n 
               + *alpha * p_n - (*++gamma) *  q_old;
    q_old   =  pq_temp;
    pq_temp =  p_n;
    p_n     = ( (*alpha) * x + (*++beta) ) * p_n - (*gamma) *  p_old;
    p_old   =  pq_temp;
  }
  sturm(mid(x), p_n);
  return 0;
}



int not_symm::sturmqr(l_real x){
 // computes changes of sign (verified) using QR decomposition
  int sign_old, sign_new, sign_changes = 1;
  l_interval *alpha = i_alpha, *beta = i_beta,  *gamma = i_gamma;   
  l_rvector y_old(2), y_new(2);
  l_ivector z_old(2), z_new(2), d_top(2);
  l_rmatrix B_old(2, 2), B_new(2, 2);
  l_imatrix A(2, 2), B_inv(2, 2);
   
  y_new[1] = mid(*alpha); 
  z_new[1] = y_new[1] - *alpha;     
  sign_old = sign(*alpha);
  if (sign_old == 2) return 0;
  y_new[2] =  mid(*++beta - *++alpha  * x) ; ++gamma;
  z_new[2] = y_new[2] + *alpha * x - *beta;
  sign_new = sign(*beta - (*alpha) * x);
  if(sign_new ==  2) return 0;
  if (sign_old * sign_new < 0){sign_changes++; sign_old = sign_new;}  
  B_new[1][1] = 1;
  B_new[1][2] = 0;
  B_new[2][1] = 0;
  B_new[2][2] = 1;
  
  A[1][1] = _l_interval(0,0);
  A[1][2] = _l_interval(1,1);

  for(int i=2; i <= grad; i++){
    A[2][1] = -*++gamma;
    A[2][2] = (*++beta) - x * (*++alpha);  
    B_old = B_new;
    y_old = y_new;
    z_old = z_new;
    y_new = mid(A) * y_old;
    d_top = y_new;
    multsub(A, y_old, d_top);
    B_new = mid(A*B_old);
    sort(B_new, z_old);
    QR(B_new);
    invert(B_new, B_inv);
    z_new = B_inv * A * B_old * z_old + B_inv * d_top;
    z_old = y_new + B_new * z_new;
    sign_new = sign(z_old[2]);
    if(sign_new ==  2) return 0;
    if (sign_old * sign_new < 0){sign_changes++; sign_old = sign_new;}  
  }
  return sign_changes;
}



int not_symm::sturmqr(l_real x, l_interval& p_n){
  // computes p_n(x) (verified)  using QR decomposition 
  
  l_interval *alpha = i_alpha, *beta = i_beta,
             *gamma = i_gamma;   
  l_rvector y_old(2), y_new(2);
  l_ivector z_old(2), z_new(2), d_top(2);
  l_rmatrix B_old(2, 2), B_new(2, 2);
  l_imatrix A(2, 2), B_inv(2, 2);
  
  y_new[1] = mid(*alpha); 
  z_new[1] = y_new[1] - *alpha;     
  y_new[2] = mid(*++alpha) * x - mid(*++beta); ++gamma;
  z_new[2] = y_new[2] - *alpha * x + *beta;
  B_new[1][1] = 1;
  B_new[1][2] = 0;
  B_new[2][1] = 0;
  B_new[2][2] = 1;
  
  A[1][1] = _l_interval(0,0);
  A[1][2] = _l_interval(1,1);

  for(int i=2; i <= grad; i++){
    A[2][1] = -*++gamma;
    A[2][2] =  x * (*++alpha) - (*++beta);  
    B_old = B_new;
    y_old = y_new;
    z_old = z_new;
    y_new = mid(A) * y_old;
    d_top = y_new;
    multsub(A, y_old, d_top);
    B_new = mid(A * B_old);
    sort(B_new, z_old);
    QR(B_new);
    invert(B_new, B_inv);
    z_new = B_inv * A * B_old * z_old + B_inv * d_top;
   }
   z_old = y_new + B_new * z_new;
   p_n = z_old[2]; 
   return 0;
}



int not_symm::sturmqr(l_interval x, l_interval& q_n, l_interval& p_n){
  // computes p_n'(x) (verified)  using QR decomposition
 
  l_interval l_p_n, alpha_i;
  l_interval *alpha = i_alpha, *beta = i_beta, 
             *gamma = i_gamma;  
  l_rvector y_old(2), y_new(2), y_deriv(2);
  l_ivector z_old(2), z_new(2), z_deriv(2), d_top(2), d_deriv(2);
  l_rmatrix B_old(2, 2), B_new(2, 2), B_deriv(2, 2);
  l_imatrix A(2, 2), B_inv(2, 2);
 
  y_new[1] = mid(*alpha);
  z_new[1] = y_new[1] - *alpha;  
  y_new[2] = mid(*++alpha * x - *++beta); ++gamma;
  z_new[2] = y_new[2] -  *alpha * x + *beta;
  B_new[1][1] = 1;
  B_new[1][2] = 0;
  B_new[2][1] = 0;
  B_new[2][2] = 1;
  
  y_deriv[1] = 0; 
  y_deriv[2] = mid(*alpha);
  B_deriv = B_new;
  z_deriv[1] = _l_interval(0,0);
  z_deriv[2] = y_deriv[2] - *alpha; 
  d_deriv[1] = 0;

  l_p_n   = y_new[2] + z_new[2];

  A[1][1] = _l_interval(0,0);
  A[1][2] = _l_interval(1,1);

  for(int i=2; i <= grad; i++){
    A[2][1] = -*++gamma;
    A[2][2] = x * *++alpha - *++beta;  

     // compute p'_i(x)!
    B_old = B_deriv;
    y_old = y_deriv;
    z_old = z_deriv;
    d_deriv[2] = *alpha *  l_p_n;
    y_deriv = mid(A) * y_old + mid(d_deriv);
    d_top   = y_deriv;
    multsub(A, y_old, d_deriv, d_top);
    B_deriv = mid(A * B_old);
    sort(B_deriv, z_old);
    QR(B_deriv);
    invert(B_deriv, B_inv);
    z_deriv = B_inv * A * B_old * z_old + B_inv * d_top;
    p_n = l_p_n; // store p_{i-1} in p_n

    // compute p_i(x)
    B_old = B_new;
    y_old = y_new;
    z_old = z_new;
    
    y_new = mid(A) * y_old;
    d_top = y_new;
    multsub(A, y_old, d_top);
    B_new = mid(A*B_old);
    sort(B_new, z_old);
    QR(B_new);
    invert(B_new, B_inv);
    z_new = B_inv * A * B_old * z_old + B_inv * d_top;
    z_old = y_new + B_new * z_new;
    l_p_n = z_old[2];    
  }
   
  z_deriv = y_deriv + B_deriv * z_deriv;
  q_n = z_deriv[2];   
  return 0;
}



// ---------------------------------------------------------------------------
// member functions weight2 (not_symm)
// ---------------------------------------------------------------------------



l_interval not_symm::weight2(l_interval& x){
 // Compute weight for node x using 
 // w  = 1 / ( sqr( p_0(x) ) / h_0 + .... + sqr( p_{n-1} ) / h_{n-1} )
 // with h_i = (p_i, p_i)

  l_interval *alpha = i_alpha, *beta = i_beta,
             *gamma = i_gamma,
             result, p_i, h_i;
 
  l_rvector y_old(2), y_new(2);
  l_ivector z_old(2), z_new(2), d_top(2);
  l_rmatrix B_old(2, 2), B_new(2, 2);
  l_imatrix A(2, 2), B_inv(2, 2);
      
  y_new[1] = mid(*alpha); 
  z_new[1] = y_new[1] - *alpha;  
  result   = *alpha;   
  y_new[2] = mid(*++alpha * x - *++beta) ; ++gamma;
  z_new[2] = y_new[2] - *alpha * x + *beta;
  B_new[1][1] = 1;
  B_new[1][2] = 0;
  B_new[2][1] = 0;
  B_new[2][2] = 1;
  
  A[1][1] = _l_interval(0,0);
  A[1][2] = _l_interval(1,1);
  p_i     = *alpha * x - *beta;

  if (orthonormal) result = sqr(i_alpha[0]) + sqr(p_i);
  else result = sqr(i_alpha[0]) / *hscalp + sqr(p_i) / hscalp[1];

  for(int i=2; i < grad; i++){ 
    A[2][1] = -*++gamma;
    A[2][2] = x * (*++alpha) - (*++beta);  
    B_old = B_new;
    y_old = y_new;
    z_old = z_new;
    y_new = mid(A) * y_old;
    d_top = y_new;
    multsub(A, y_old, d_top);
    B_new = mid(A*B_old);
    sort(B_new, z_old);
    QR(B_new);
    invert(B_new, B_inv);
    z_new = B_inv * A * B_old * z_old + B_inv * d_top;
    z_old = y_new + B_new * z_new;
    p_i = z_old[2];
    if (orthonormal) result += sqr(p_i);
    else result += sqr(p_i) / hscalp[i];
    }
  if (0 <= result) throw weight_error();

  // cout << " weight  " << stagprec << endl; 


  return 1.0 / result; 
}



// ---------------------------------------------------------------------------
// member functions (test functions) 
// --------------------------------------------------------------------------- 


void not_symm::test_coeff(ostream& os){
  l_real      *pr_alpha = r_alpha, *pr_beta = r_beta, *pr_gamma = r_gamma;
  l_interval  *pi_alpha = i_alpha, *pi_beta = i_beta, *pi_gamma = i_gamma;
  os << " Eingabe von stagprec ";
  cin >> stagprec;
  calc_coefficients();
  cout << " stagprec = " << stagprec << endl;
  for(int i=0; i<=grad; i++)
    os // << " i_alpha[" << i << "] " << pi_alpha[i] << " diam " 
      //  << diam(i_alpha[i]) << endl << " i_beta[" << i << "] " 
      << i  << " & " // pi_beta[i] << " diam " 
      << diam(i_beta[i]) << " & "  
      //  << " i_gamma[" << i 
      // << "] " << pi_gamma[i] << " diam " 
      << diam(i_gamma[i]) << " & " << endl;
  //       << " r_alpha[" << i << "] " << pr_alpha[i] << endl
  //    << " r_beta[" << i << "] " << pr_beta[i] << endl   
  //    << " r_gamma[" << i << "] " << pr_gamma[i] << endl << endl; 
}


// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// derived classes (not_symm)
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------



// ---------------------------------------------------------------------------
// class coeff (not_symm)
// ---------------------------------------------------------------------------

coeff::coeff(int n,real min,real max, init_function alpha, init_function beta,
             init_function gamma, l_interval h0):not_symm(n)
{
  n_max          = n;
  inf            =  min;
  sup            =  max;
  if (h0 == 0)
    orthonormal  = true;
  else{
    orthonormal  = false;
    hscalp       = new l_interval[n];
    *hscalp      = h0;
  }   
  
  init_alpha     = alpha;
  init_beta      = beta,
  init_gamma     = gamma;
}



coeff::~coeff(){
  if(!orthonormal) delete[] hscalp;
}



int coeff::calc_coefficients(){

  l_interval *pi_alpha = i_alpha, *pi_beta = i_beta, *pi_gamma = i_gamma; 
  l_real     *pr_alpha = r_alpha, *pr_beta = r_beta, *pr_gamma = r_gamma; 


  *pi_alpha = init_alpha(0);
  *pr_alpha = mid(*pi_alpha);
  *pi_beta  = 0;
  *pr_beta  = 0;
  *pi_gamma = 0;
  *pr_gamma = 0;  
  for(int i = 1 ; i <= grad ; i++){
     *++pi_alpha = init_alpha(i);
     *++pr_alpha = mid(*pi_alpha);
     *++pi_beta  = init_beta(i);
     *++pr_beta  = mid(*pi_beta);
     *++pi_gamma = init_gamma(i);
     *++pr_gamma = mid(*pi_gamma);
  }
  calc_scalps(false);
  return stagprec;
}



// ---------------------------------------------------------------------------
// class gautschi_symm
// ---------------------------------------------------------------------------

gautschi::gautschi(int n, real min, real max, init_function calc_moments)
 :not_symm(n){
 //  generic type with init functions for moments
 
  n_max          = n;
  sup            = max;
  inf            = min;
  orthonormal    = true;
  init_moment    = calc_moments;
}



int gautschi::calc_coefficients(){
 // compute coefficients from ordinary moments

  int n = 2* grad;
  idotprecision dot;
  l_interval*  c = new l_interval[n + 1];
  l_interval**  r = new l_interval*[grad + 1]; // (r)_ij = R transposed
  for(int i=0; i<=n;i++) c[i] = init_moment(i);
  for(int i=0; i<=grad;i++) r[i] = new l_interval[i + 1];
  r[0][0] = sqrt(c[0]); // j = 0    
  for(int j=1; j<=grad; j++){
    for(int i=0; i<j; i++){
      dot = c[i+j];
      for(int k=0; k<i; k++)
        accumulate(dot, -r[i][k], r[j][k]);
      r[j][i] = dot;
      r[j][i] /= r[i][i];
    }
    dot = c[2*j];
    for(int k=0; k<j; k++)
      accumulate(dot, -r[j][k], r[j][k]);
      r[j][j] = dot;
      r[j][j] = sqrt(r[j][j]);
  }

  l_interval *pi_alpha = i_alpha, *pi_beta = i_beta, *pi_gamma = i_gamma; 
  l_real     *pr_alpha = r_alpha, *pr_beta = r_beta, *pr_gamma = r_gamma; 

  *pi_alpha = 1 / r[0][0];
  *pr_alpha = mid(*pi_alpha);
  *pi_beta  = 0; // not used
  *pr_beta  = 0; // not used
  *pi_gamma = 0; // not used
  *pr_gamma = 0; // not used
  *++pi_alpha = 1 / r[1][1] ; // p_1 = alpha_1 * x 
  *++pr_alpha = mid(*pi_alpha);
  *++pi_beta  = r[1][0] / r[1][1] / r[0][0]; 
  *++pr_beta  = mid(*pi_beta) ; 
  *++pi_gamma = 0; // not used;
  *++pr_gamma = 0; // not used;
  for(int i = 2 ; i <= grad ; i++){
    *++pi_alpha = r[i-1][i-1] / r[i][i];
    *++pr_alpha = mid(*pi_alpha);
    *++pi_beta = *pi_alpha * ( r[i][i-1]   / r[i-1][i-1] 
                               - r[i-1][i-2] / r[i-2][i-2] );
    *++pr_beta = mid(*pi_beta);
    *++pi_gamma = sqr( r[i-1][i-1] ) / ( r[i][i] * r[i-2][i-2] );
    *++pr_gamma = mid(*pi_gamma);
  } 
  for(int i=0; i<=grad; i++) delete[] r[i];
  delete[] r;
  return Actual_Stagprec(*pi_alpha); 
}



// ---------------------------------------------------------------------------
// class sado (not_symm)
// ---------------------------------------------------------------------------

sado::sado(int n, real min, real max, init_function alpha, init_function beta,
     init_function gamma,  init_function calc_moments, bool calc_m):not_symm(n)
{
 // generic type for modified moments

  sup             = max;
  inf             = min;
  n_max           = n;
  orthonormal     = false;
  ordinary        = calc_m;
  init_alpha      = alpha;
  init_beta       = beta;
  init_gamma      = gamma;
  init_moment     = calc_moments;
  hscalp          = new l_interval[n];
  
}



sado::~sado(){
  delete[] hscalp;
}



void sado::calc_mod(l_interval** s, l_interval* alpha_mod, l_interval* beta_mod
                                   , l_interval* gamma_mod)
{
 //  computes the modified moments from the ordinary moments
 //  and the start values for the Sack/Donovan iteration

  stagprec++;

  int i,j;
  l_interval** const nu = new l_interval*[2*grad];

  nu[0] = new l_interval[2 * grad];
  for(j=0; j < 2*grad; j++) nu[0][j] = init_moment(j);
  s[0][0] = 1;
  nu[1] = new l_interval[2 * grad - 1];

  for(j=0; j < 2*grad-1; j++) 
    nu[1][j] = alpha_mod[1] * nu[0][j+1] - beta_mod[1] * nu[0][j]; 
  s[0][1] = nu[1][0] / nu[0][0];
  
  for(i=2; i < 2*grad; i++){
    nu[i] = new l_interval[2 * grad  - i];
    for(j = 0; j < 2 * grad - i; j++)
      nu[i][j] =  alpha_mod[i] * nu[i-1][j+1] - beta_mod[i] * nu[i-1][j] 
                                             - gamma_mod[i] * nu[i-2][j];
    s[0][i] = nu[i][0] / nu[0][0];  
  }

  for(i=0; i < 2*grad; i++) delete[]  nu[i];

  // cout << " sado  "  << stagprec << endl; 

  delete[] nu;
  stagprec-- ;
}



int sado::calc_coefficients(){
 // Sack/Donovan iteration:
 // compute the coefficients from the modified moments
 // and coefficients


  int i, j, n = grad;

  l_real     *pr_alpha  = r_alpha,*pr_beta = r_beta, *pr_gamma = r_gamma; 
  l_interval *pi_alpha  = i_alpha, *pi_beta = i_beta, *pi_gamma = i_gamma;
  l_interval sigma, rho;

  l_interval *alpha_mod = new l_interval[2*n], 
            *beta_mod  = new l_interval[2*n],
            *gamma_mod = new l_interval[2*n];

  l_interval** const s = new l_interval*[n];

 // memory allocation rows s[i]
  for(j=0; j<n; j++)
    s[j] = new l_interval[2 * grad  - j]; 
 
 // init coefficients alpha__mod, beta_mod, gamma_mod
  for(i=0; i<=2*n-1;i++){
    alpha_mod[i] = init_alpha(i);
    beta_mod[i] =  init_beta(i);
    gamma_mod[i] = init_gamma(i);
  }

 // store coefficients alpha_0, beta_0, gamma_0, gamma_1
  *pi_alpha   = 1.0;
  *pr_alpha   = 1.0;
  *pi_beta    = 0;
  *pr_beta    = 0;
  *pi_gamma   = 0; 
  *pr_gamma   = 0;
  *++pi_gamma = 0; 
  *++pr_gamma = 0;

  if(ordinary){ 
    calc_mod(s, alpha_mod, beta_mod, gamma_mod);
    *hscalp    = init_moment(0);
  }
  else{    //  compute the modified moments using the init function!
    *hscalp  =  init_moment(0);
    *hscalp /= alpha_mod[0];
  for(j=0; j <= 2 * n - 1; j++)
    s[0][j] = init_moment(j) / *hscalp;
  }

  sigma = ( s[0][1] + beta_mod[1] ) / alpha_mod[1];
  rho   = ( beta_mod[2] /  alpha_mod[2] - sigma ) * s[0][1] +
          ( s[0][2] + gamma_mod[2] ) / alpha_mod[2]; 

   // store coefficients alpha_1, beta_1, gamma_2,
  *++pi_alpha = 1.0;
  *++pr_alpha = 1.0;
  *++pi_beta  = sigma;
  *++pr_beta  = mid(sigma);
  *++pi_gamma = rho / alpha_mod[1]; 
  *++pr_gamma = mid(*pi_gamma);
     
   // compute s[1][*]
  s[1][0] = ( (beta_mod[1] / alpha_mod[1] - sigma) * s[0][0] +
               s[0][1] / alpha_mod[1] ) / rho;
  for(j=1; j <= 2 * n - 2; j++){
    s[1][j] =( (beta_mod[j+1] / alpha_mod[j+1] - sigma) * s[0][j] +
             ( s[0][j+1] + gamma_mod[j+1] * s[0][j-1]) / alpha_mod[j+1] ) / rho;
  }
  for(i=1 ; i <= n - 2; i++){   
   // compute sigma_i, rho_i;
  sigma = ( s[i][i+1] + beta_mod[i+1] ) / alpha_mod[i+1] - s[i-1][i]
            / alpha_mod[i];
  rho   = ( beta_mod[i+2] /  alpha_mod[i+2] - sigma ) * s[i][i+1] +
          ( s[i][i+2] + gamma_mod[i+2] ) / alpha_mod[i+2] - 
            s[i-1][i+1] / alpha_mod[i];  

   // store coefficients alpha_i+1, beta_i+1, gamma_i+2,
  *++pi_alpha = 1.0;
  *++pr_alpha = 1.0;
  *++pi_beta  = sigma;
  *++pr_beta  = mid(sigma);
  *++pi_gamma = rho / alpha_mod[i+1]; 
  *++pr_gamma = mid(*pi_gamma);     

 // compute s[i+1][*]
  s[i+1][0] =( (beta_mod[1] / alpha_mod[1] - sigma) * s[i][0] +
                s[i][1]  / alpha_mod[1] -  s[i-1][0] / alpha_mod[i] ) / rho;
  for(j=1; j <= 2 * n - 2 - i; j++)
    s[i+1][j] =( (beta_mod[j+1] / alpha_mod[j+1] - sigma) * s[i][j] +
               (s[i][j+1]  + gamma_mod[j+1] * s[i][j-1]) / alpha_mod[j+1]-
                s[i-1][j] / alpha_mod[i] ) / rho;
  }

  // store alpha_n, beta_n 
  *++pi_alpha = 1.0;
  *++pr_alpha = 1.0;
  *++pi_beta  = ( s[n-1][n] + beta_mod[n] ) / alpha_mod[n] - 
                  s[n-2][n-1] / alpha_mod[n-1];
  *++pr_beta  = mid(*pi_beta); 

  delete[] alpha_mod;
  delete[] beta_mod;
  delete[] gamma_mod;
  for(i=0; i<n; i++) delete[]  s[i];
  delete[] s;
  // cout << "stagprec " << stagprec << endl;
  calc_scalps(false);
  return Actual_Stagprec(*pi_gamma); 
}













