 //============================================================================
//
//                              Program/Module
//                                   from
//                C++ package for verified numerical integration
//                  of singular and strong singular functions
//                      Copyright (c) 2000 Stefan Wedner
//
// This program/module is free software for non-commercial use. 
//
//
//============================================================================
//----------------------------------------------------------------------------
//
// File:     node.h
// See also: node.C
// Purpose:  computing the nodes and weights of gaussian  quadrature formulae
//           (with at least two nodes) using the recurrence relation:
//     alpha_0 = p_0(x) ,   
//     p_1(x)  = alpha_1 * x - beta_1  not (alpha_1 *  x - beta_1) * p_0(x) !
//     n > 1:
//     p_n(x)  = ( alpha_n x - beta_n ) p_n-1(x) - gamma_n * p_n-2(x)
//
// ===========================================================================

#ifndef _CXSC_LIMATRIX_HPP_INCLUDED  
#include "l_imatrix.hpp"
#endif

#ifndef l_rmath_hpp_included
#include "l_rmath.hpp"
#endif

#ifndef rmath_included
#include "rmath.hpp"
#endif

using namespace cxsc;
using namespace std;    

// ----------------------------------------------------------------------------
// constants:
// ----------------------------------------------------------------------------

const int precision_coef_corr = 2;
const int precision_real_corr = 2;
const int iteration_corr      = 60;
const int iteration_step      = 10;
const int max_newton          = 10;



// ----------------------------------------------------------------------------
// userdefined types: 
// ----------------------------------------------------------------------------

typedef   l_interval  (*init_function)(int);

enum polyname  { Legendre, Jacobi, Laguerre, Hermite, Tschebycheff, 
                 Generic, Generic_c, Generic_m}; 
enum outformat { Hexa, Long, Decimal, LaTex, Nodes, Weights};



// ----------------------------------------------------------------------------
// auxiliary functions:
// ----------------------------------------------------------------------------

/* l_rmatrix mid(l_imatrix A); */

/* l_rvector mid(l_ivector b); */

void QR(l_rmatrix& A);

void sort(l_rmatrix& B, l_ivector& z);

void invert(l_rmatrix& B, l_imatrix& Binv);

inline int sign(l_interval p){
 if (Inf(p) > 0.0) return  1;
 else if (Sup(p) < 0.0) return -1;
      else if (0.0 < p) return 2;
 return 0;
}
 
void multsub(l_imatrix& A, l_rvector& y_old, l_ivector& y_new);

void multsub(l_imatrix& A, l_rvector& y_old, l_ivector& d, l_ivector& y_new);

int Actual_Stagprec(l_interval& a);



class peano_base;

// ----------------------------------------------------------------------------
// class sym:
// ----------------------------------------------------------------------------

class symm{
 
public:

void decrease_grad(int j){ grad=j;}

void calc_coefficients(int precision);
// computing the coefficients of the reccurence relation
// with the given precision

int calc_node(l_interval& enclosure, int i);
// computing the i-th node
// input: enclosure of the i-th node (var enclosure), i 
// output: improved enclosure (var enclosure)
// returns precision needed for verifikation

int calc_node_and_weight(l_interval& enclosure, int i, int prec_coef =2, 
       int prec_real = 2, int iter = 50, ostream& os = cout);
// computing the i-th node and weight
// input: enclosure of the i-th node (var enclosure), i
//        and precisions for: calculating coefficients  prec_coef
//                            real bisection            prec_real
//        number of iteration (bisection)               iter                   

void calc_nodes_and_weights(const l_interval& start, int i, int j, 
                                            ostream& = cout, istream& = cin  );
//  computing nodes and weights from i to j (interactive version)
//  input: start enclosure, which is containing all nodes,
//         i,j
//     

int calc_nodes_and_weights(int prec_coef = 2, int prec_real = 2,int iter = 20,
                           ostream& os = cout);
// computing all nodes and weights

int verify(l_interval x, int k, int  precision, int prec_max);
// verifying, that the k-the node is lying in x 
// if the start precision (var precision) isn't high enough
// the precision will be increased. The maximum precision is prec_max


l_interval test_nodes_and_weights(int n = 1, int precision = 2);
// computing the quadraturesum of f(x) = power(x,n)
// for testing 


// test functions
virtual void test_coeff(ostream& os = cout);
void test_sturm(ostream& os = cout);


// output-operators
friend ostream& operator<<(ostream& os,const symm&  out); 
friend ostream& operator<<(ostream& os,const peano_base&  out);
friend ostream& operator<<(ostream& os,outformat out);

protected:
  int grad;
  int n_max;
  real sup;
  real inf;
  bool orthonormal;

  l_interval* i_alpha;   // enclosures of the coefficients alpha_i 
  l_interval* i_gamma;   // enclosures of the coefficients gamma_i  
  l_real* r_alpha;       // approximations of the coefficients alpha_i 
  l_real* r_gamma;       // approximations of the coefficients gamma_i 
  l_interval* hscalp;  


private:

  int iteration;       // number of iterations (bisect)
  int precision_coef;  // precision for computing the coefficients
  int precision_real;  // precision for calculating approximations
  int precision_ival;  // precision for calculating nodes and weights

  l_interval* nodes;     // enclosures of the nodes
  l_interval* weights;   // enclosures of the weights

  static outformat   oflag1;  
  static outformat   oflag2;


  virtual int calc_coefficients()=0;
  // computing coefficients

protected:
  symm(int n);
  ~symm();

private:
  virtual int sturm(l_real  x);
  // computes #changes of sign (not verified)
  virtual int sturm(l_interval x);
  // computes #changes of sign (verified)
  virtual int sturm(l_real x, l_interval& p_n);
  // computes p_n verified
  virtual int sturm(l_interval x, l_interval& p, l_interval& q);
  // computes p_(mid(x)) and p_n'(x) (for INewton)
  virtual int sturmqr(l_real x);
  // computes #changes of sign (verified using QR decomposition)
  virtual int sturmqr(l_real x, l_interval& p_n);
  //  computes  p_n(x)         (verified using QR decomposition)
  virtual int sturmqr(l_interval x, l_interval& q, l_interval& p);
  // computes p_n'(x) (for INewton) and p_{n-1}
  l_interval weight1(l_interval& node);
  // computes a_n * (pn-1, pn-1) / ( pn(x) * pn-1(x) )
  virtual l_interval weight2(l_interval& x);
  // computes 1 / ( pi0(x) * pi0(x) + ... + pin-1(x) * pin-1(x) )
  virtual l_interval h(int i);
  // computes sclarproduct (pn, pn)

protected:
  void calc_scalps(bool use_function);
   // computes (p0, p0) ... ( pn2, pn2)

private:
  int verify(l_interval x, int k, int precision);
  // verifies " k-th root of p_n is in x "
  void bisect(l_interval& start, int k);
  // bisects start iteration times
  int  newton(l_interval& node);
  // inewton step
  int calc_node_and_weight(int, int, int, l_interval&, int, ostream&);
  // computing node and weight 
};



class legendre: public symm{
 public:
   legendre(int n);  
   ~legendre();  
   
 private:
   int calc_coefficients();
   l_interval weight(l_interval& node);
   l_interval h(int i);

};



class coeff_symm: public symm{
 public:
   coeff_symm(int n, real max, init_function alpha, 
              init_function gamma, l_interval h0);
   ~coeff_symm();  
 private:
   init_function  init_alpha;   // calculates coeff. alpha_i
   init_function  init_gamma;   // calculates coeff. gamma_i       
   int calc_coefficients();
};



class gautschi_symm: public symm{
 public:
   gautschi_symm(int n, real max, init_function calc_moments);
 private:
   init_function init_moment;  // calculates moments 
   int calc_coefficients();
};



class sado_symm: public symm{
 public:
   sado_symm(int n, real max, init_function alpha, init_function gamma,
                         init_function calc_moments, bool calc_m);
   ~sado_symm();  
 private:
   bool ordinary;                // ordinary or mod. moments
   init_function  init_moment;  // calculates moments
   init_function  init_alpha;   // calculates mod. coeff. alpha_i
   init_function  init_gamma;   // calculates mod. coeff. gamma_i
   int calc_coefficients();
   void calc_mod(l_interval**, l_interval*, l_interval*);
};



class new_classic_symm: public symm{
 public:
   new_classic_symm(int n);
   ~new_classic_symm();  
  // you have to set h_0 (p0,p0) or
  // define (overwrite)  void h(int i);
 private:
   int calc_coefficients();
  //  void h(int i);
};



class not_symm:public symm{

public: 

void test_coeff(ostream& os = cout);
// for testing

protected:

  l_interval* i_beta;   // enclosures of the coefficients beta_i  
  l_real* r_beta;       // approximations of the coefficients beta_i 

  not_symm(int n);
  // constructor
  ~not_symm();
  // destructor

private:

  int sturm(l_real  x);
  // computes #changes of sign (not verified)
  int sturm(l_interval x);
  // computes #changes of sign (verified)
  int sturm(l_real x, l_interval& p_n);
  // computes p_n verified
  int sturm(l_interval x, l_interval& p, l_interval& q);
  // computes p_(mid(x)) and p_n'(x) (for INewton)
  int sturmqr(l_real x);
  // computes #changes of sign (verified using QR decomposition)
  int sturmqr(l_real x, l_interval& p_n);
  //  computes  p_n(x)         (verified using QR decomposition)
  int sturmqr(l_interval x, l_interval& q, l_interval& p);
  // computes p_n'(x) (for INewton) and p_{n-1}
  l_interval weight2(l_interval& x);
  // computes 
  virtual int calc_coefficients()=0;
};



class coeff: public not_symm{
 public:
   coeff(int n, real min, real max, init_function alpha, 
         init_function beta, init_function gamma, l_interval h0);
   ~coeff();  
 private:
   init_function  init_alpha;   // calculates coeff.  alpha_i
   init_function  init_beta;    //  calculates coeff. beta_i       
   init_function  init_gamma;   //  calculates coeff. gamma_i  
   int calc_coefficients();
};



class gautschi: public not_symm{
 public:
   gautschi(int n, real min, real max, init_function calc_moments);
 private:
   init_function init_moment;  // calculates moments 
   int calc_coefficients();
};



class sado: public not_symm{
 public:
   sado(int n, real min, real max, init_function alpha,init_function beta,
        init_function gamma, init_function calc_moments, bool calc_m = false);
   ~sado();  

 private:
   bool ordinary;               // ordinary or mod. moments
   init_function  init_moment;  // calculates moments
   init_function  init_alpha;   // calculates mod. coeff. alpha_i
   init_function  init_beta;    // calculates mod  coeff. beta_i   
   init_function  init_gamma;   // calculates mod. coeff. gamma_i
   int calc_coefficients();
   void calc_mod(l_interval**, l_interval*,  l_interval*, l_interval*);
};



class new_classic: public not_symm{
 public:
   new_classic(int n);
   ~new_classic();  
  // you have to set h_0 (p0,p0) or
  // define (overwrite)  void h(int i);
 private:
   int calc_coefficients();
  //  void h(int i);
};



// ----------------------------------------------------------------------------
// error classes
// ----------------------------------------------------------------------------

struct newton_error{
};

struct weight_error{
};

struct verify_error{
 int i;
 verify_error(int j):i(j){}
};

struct qr_error{
};

struct invert_error{
};

struct fatal_error{
};

struct incomplete_error{
 int i;
 incomplete_error(int j):i(j){}
};

