//============================================================================
//
//                              Program/Module
//                                   from
//                C++ package for verified numerical integration
//                  of singular and strong singular functions
//                      Copyright (c) 2000 Stefan Wedner
//
// This program/module is free software for non-commercial use. For details
// on theory, algorithms, and programs, see:
//
//
//============================================================================
//----------------------------------------------------------------------------
//
// File:     peano.C
// See also: peano.h, peano.tpl
// Purpose:  computing the peano constants of quadrature formulas
//     
// ---------------------------------------------------------------------------

#include "peano.h"
#include <iostream> 

int  peano_base::start_index;
idotprecision peano_base::i_dot;
real peano_base::integrate_from;
real peano_base::integrate_to;


inline int inside (const real& x, const interval& y ){      
// Contained-in-the-interior relation
  return  (Inf(y) < x) && (x < Sup(y)) ;
} 

   
inline bool nonempty_intersection( interval a, interval b ){
// returns true, if intersection is not empty
  return Sup(a)>=Inf(b) && Inf(a)<=Sup(b);
}


// ----------------------------------------------------------------------------
// abstract class peano_base member functions
// ----------------------------------------------------------------------------


// ---------------------------constructor -------------------------------------

peano_base::peano_base(string filename, bool symm){ 

  int  i;
  real a_down, a_up;
  ifstream external_data(filename.data());

  if(! external_data) cout << " error! : file " << filename 
                           << " not available! " << endl;
  external_data >> number_nodes;

  weight        = new interval[number_nodes+1];
  node          = new interval[number_nodes+1];
  symmetric     = symm;
  positive_part = 0;
  negative_part = 0;

  // symmetric case:
  if (symmetric){
    // reading nodes from file
    for(i=1; i<=number_nodes/2; i++){
      external_data >> Hex >>  a_down;
      external_data >> Hex >>  a_up;
      node[i] = interval(a_down,a_up);
      node[number_nodes-i+1] = -  node[i]; 
    }
    if(number_nodes%2){
      external_data >> Hex >> a_down;
      external_data >> Hex >> a_up;
      node[number_nodes/2 + 1] = interval(a_down,a_up);
    } 
    // reading weights from file
    for(i=1; i<=number_nodes/2; i++){
      external_data >> Hex >> a_down;
      external_data >> Hex >> a_up;
      weight[i] = weight[number_nodes-i+1] = interval(a_down,a_up); 
    }
    if(number_nodes%2){
       external_data >> Hex >> a_down;
       external_data >> Hex >> a_up;
       weight[number_nodes/2 + 1] = interval(a_down,a_up);
    }
  }
  // nonsymmetric case:
  else{
    // reading nodes from file
    for(i=1; i<=number_nodes; i++){
      external_data >> Hex >> a_down;
      external_data >> Hex >> a_up;
      node[i] = interval(a_down,a_up); 
    }
    // reading weights from file
    for(i=1; i<=number_nodes; i++){
      external_data >> Hex >> a_down;
      external_data >> Hex >> a_up;
      weight[i] =  interval(a_down,a_up);     
    }
  }
  external_data.close();
  cout << Dec;    
}


peano_base::peano_base(){ 
  positive_part = 0;
  negative_part = 0;
}



// ----- member function calculate --------------------------------------------

void peano_base::calculate(int s){

  bool     left_closed = false, right_closed = false;
  int      j, help;
  interval kernel_node;

  order_kernel = s;
    
  positive_part = 0;
  negative_part = 0;

  init_data();

  if (symmetric) start_index = help = (number_nodes + 2) /2;
  else           start_index = help = 1;   

  // integration over intervals node[1], .... , node[n]

  if (Inf(node[start_index]) <= 0){ 
     left_closed = true; 
     integrate_from = Sup(node[start_index]);
     j = start_index;   
     help = ++start_index;
     integrate_rectangle(interval(0,Sup(node[j])), kernel(j));
  }
  else integrate_from = 0;

  for( ; start_index< number_nodes; start_index++ ){
     //    integrate_rectangle(node[start_index], kernel(node[start_index]));
       kernel_node = kernel(start_index); 
       kernel_node = kernel_node | kernel(start_index--); 
       integrate_rectangle(node[++start_index], kernel_node);
  }    
  if (Sup(node[start_index]) >= 1){     
    integrate_rectangle(node[start_index], kernel(start_index));
    right_closed = true;
  }
  else{
    kernel_node = kernel(start_index); 
    kernel_node = kernel_node | kernel(start_index--); 
    integrate_rectangle(node[++start_index], kernel_node);
  }


  // integration over intervals [Sup(node[j]), Inf(node[j+1]) ]
  start_index = help; 

  // not symm and definite
  if (!symmetric) {
     integrate_to    = Inf(node[start_index]);
     integrate_stamm(integrate_from, integrate_to);
     integrate_from  = Sup(node[start_index]);
     help++; 
  }

  for(start_index = help ; start_index <= number_nodes; start_index++ ){
    integrate_to    = Inf(node[start_index]); 
    check_closer();
    integrate_from  = Sup(node[start_index]);         
  }
  // not definite -> replace integrate_stamm by check_closer
  if (!right_closed){
    integrate_to    = 1; 
    integrate_stamm(integrate_from, integrate_to);
  }
  use_symmetrie();     
} 


// ------------------ member function check_closer ----------------------------

void peano_base::check_closer(){

   if ( !inside(0, mvf_kernel( interval(integrate_from, integrate_to))) )
      integrate_stamm(integrate_from, integrate_to);
   else{
       extented_newton(interval(integrate_from, integrate_to));
       if( integrate_from!= integrate_to ) 
               integrate_stamm(integrate_from, integrate_to);
  }
}



// -------------- meber function integrate_rectangle --------------------------

void peano_base::integrate_rectangle(const interval& x, 
                                     const interval& kernel_x){
  interval d_x; 
  i_dot = 0; 
  accumulate(i_dot, -Inf(x), order_kernel);
  accumulate(i_dot,  Sup(x), order_kernel );
  rnd(i_dot, d_x);

  //  cout << " rectaival " << x << endl << " kernel_x " << kernel_x << endl
  //     << " rectangle " << d_x * kernel_x << endl;

  if( inside(0, kernel_x) ) {
     accumulate(positive_part,d_x, interval(0, Sup(kernel_x)));
    // positive part
     accumulate(negative_part,d_x, interval(Inf(kernel_x), 0));
    // negative part
  } 
  else if( Inf(kernel_x)>=0 ) 
          accumulate(positive_part, d_x, kernel_x );
       else                   
          accumulate(negative_part, d_x, kernel_x ); 
}



// -------------- meber function extented_newton ------------------------------

void peano_base::extented_newton(interval x){
  //cout << "ext " << x << endl << mvf_kernel(x) << " d " << diam(kernel(x)) << endl;
  // extented interval newton algorithm  
  if( diam(x)<=epsilon ) {               // stop recursion
    if( integrate_from < Inf(x) )          
     integrate_stamm(integrate_from, Inf(x));
    integrate_rectangle(x, mvf_kernel(x));  
    integrate_from = Sup(x);                   
  } 
  else {                                // next recursion step
    real midx           = mid(x);                      
    interval K_midx     = kernel( interval(midx,midx) );
    interval diffK_x    = mvf_derivation1(x) ;

    interval n_x;

    if( Sup(diffK_x) < 0 || Inf(diffK_x) > 0){   
    // Newton step
      n_x = midx - K_midx  / diffK_x;
      if( Inf(x)<=Sup(n_x) && Sup(x)>=Inf(n_x) ){ 
      // else no root! 
	n_x = n_x & x; 
	if( n_x == x ){         // bisection 
	  extented_newton( interval( Inf(x), midx ) );  
	  extented_newton( interval( midx , Sup(x)) );
	} 
        else  extented_newton(n_x); 
      } 
    } 
    else{     // extented division 
      
      if( Inf(K_midx)>0 ){      
        if( Sup(diffK_x)!=0 ){ 
          n_x = -MaxReal | ( midx - K_midx / Sup(diffK_x) );
	  if(nonempty_intersection(n_x, x)) extented_newton(n_x & x); 
        }
	if ( Inf(diffK_x)!=0 ){
          n_x =  MaxReal | ( midx - K_midx / Inf(diffK_x) );
	  if(nonempty_intersection(n_x, x)) extented_newton(n_x & x); 
	}
      }
      else if( Sup(K_midx)<0 ){                             
        if( Inf(diffK_x)!=0 ){
          n_x =  -MaxReal | ( midx - K_midx / Inf(diffK_x) );
	  if(nonempty_intersection(n_x, x)) extented_newton(n_x & x); 
	} 
	if( Sup(diffK_x)!=0 ){
          n_x =  MaxReal | ( midx - K_midx / Sup(diffK_x) );
	  if(nonempty_intersection(n_x, x)) extented_newton(n_x & x); 
	}
      } 
      else {  // no improvement -> bisection
              extented_newton(interval(Inf(x), midx));
              extented_newton(interval(midx, Sup(x)));
      }            
    }
  }
}



// ------------------ meber function mvf_derivation1 --------------------------

interval peano_base::mvf_derivation1(interval x){

  interval result;
  
  if( diam(x)<1) {
    interval mid_x; mid_x = mid(x);    // new C-XSC  !!! remove
    interval diffK_x = derivation1(mid_x);
    mid_x = x - mid_x;
    interval diff2K_x = derivation2(x);

    i_dot = diffK_x;
    accumulate(i_dot, diff2K_x, mid_x );
    rnd( i_dot, result );
    return result;
  } 
  else return derivation1(x);
}



// --------------------- member function mvf_kernel ---------------------------

interval peano_base::mvf_kernel(interval x){

 interval result;
  if( diam(x)<1) {
    interval mid_x; mid_x = mid(x); // new C-XSC  !!! remove
    interval K_mid_x = kernel(mid_x);
    mid_x = x - mid_x;
    interval diffK_x; diffK_x = derivation1(x);
   //  interval diffK_x = mvf_derivation1(x);
    i_dot =  K_mid_x;
    accumulate( i_dot, diffK_x, mid_x);
    rnd( i_dot, result );
    return result;
  } else
    return kernel(x);
}



// --------------------- friend function << operator --------------------------

ostream& operator<<(ostream& os, const peano_base&  out){

 interval  negative, positive;

 rnd(out.negative_part, negative);
 rnd(out.positive_part, positive);
 
 if (symm::oflag1 == Hexa){
   os << Hex << Inf(negative) << endl << Sup(negative) << endl
      << Inf(positive) << endl << Sup(positive) << endl  << Dec; 
 }
 else if (symm::oflag1 == Decimal){
   os << Dec <<  SetPrecision(26,20) << negative << endl << positive << endl; 
 } 
  return os;
}



// ---------------------------------------------------------------------------
// member functions class peano
// ---------------------------------------------------------------------------


// --------------------- member functions kernel ------------------------------

interval peano::kernel(int k){
   // special case: intervals [x_i]
   // enclosure of (s-1)! * k_s(x)
   interval result, diff, x = node[k];
  
   i_dot = power(1 - x, order_kernel) / order_kernel;

   for(int  i= start_index; i<=number_nodes; i++ ){
     diff = truncated_diff(node[i],x);
     if( diff!=0 ) // diff = 0 only possible if x_i are point intervals
       if( order_kernel == 1)
         if (Inf(node[i])<Sup(x)) diff = interval(0,1);
         else diff = 1;
       else diff = power(diff, order_kernel - 1);
     accumulate(i_dot, -weight[i], diff);
   }
   rnd( i_dot, result );
   return result;
}

interval peano::kernel(interval x){
   // enclosure of (s-1)! * k_s(x)
   interval result, diff;
  
   i_dot = power(1 - x, order_kernel) / order_kernel;

   for(int  i= start_index; i<=number_nodes; i++ ){
     if( order_kernel==1) diff = 1;
     else diff = power( node[i] - x, order_kernel - 1);
     accumulate(i_dot, -weight[i], diff);
   }
   rnd( i_dot, result);
   return result;
}


// --------------------- member function derivation1 --------------------------

interval peano::derivation1(interval x){

  interval result, diff;
  
  i_dot = -power(1 - x, order_kernel-1) ;

   for(int  i= start_index; i<=number_nodes; i++ ){
     if( order_kernel == 1 )      diff = 0;
     else if ( order_kernel == 2) diff = 1;
     else diff = (order_kernel - 1)*power(node[i]-x, order_kernel - 2);
     accumulate(i_dot, weight[i], diff);
   }
   rnd( i_dot, result );
   return result;
}


// --------------------- member function derivation2 --------------------------

interval peano::derivation2(interval x){
 interval result, diff;
  
   i_dot = power(1 - x, order_kernel - 2) * (order_kernel - 1);

   for(int  i= start_index; i<=number_nodes; i++ ){
     if( order_kernel < 3 ) diff = 0;
     else if ( order_kernel == 3) diff = 2;
          else diff = (order_kernel-2) * (order_kernel-1)*
                       power(node[i]-x, order_kernel-3);
     accumulate(i_dot, -weight[i], diff);
   }
   rnd( i_dot, result );
   return result;
}



// --------------------- member integrate_stamm -------------------------------

void peano::integrate_stamm(real from, real to){

  interval diff; 
  i_dot = truncated_diff( power(1.0 -interval(from,from), order_kernel+1), 
                          power(1.0 -interval(to,to), order_kernel+1) ) 
                        / (order_kernel+1);
  for( int i=start_index; i<=number_nodes; i++ ) {  
    diff = truncated_diff( power(node[i]-from, order_kernel), 
                          power(node[i]-to, order_kernel));
    accumulate( i_dot, -weight[i], diff);
  }

  rnd( i_dot, diff );
  if( Inf(diff)>= 0 )
    positive_part += i_dot;
  else if( Sup(diff)<=0 ) negative_part += i_dot;
  else {
    positive_part +=  interval(0,Sup(diff));
    negative_part +=  interval(Inf(diff),0); 
  }
}


// --------------------- member function use_symmetrie ------------------------

void peano::use_symmetrie(){
  if (odd(order_kernel)){
    i_dot = positive_part;
    positive_part -= negative_part;
    negative_part -= i_dot;
  }
  else {
      positive_part += positive_part; 
      negative_part += negative_part;
  }

}


