//============================================================================
//
//                              Program/Module
//                                   from
//                C++ package for verified numerical integration
//                  of singular and strong singular functions
//                      Copyright (c) 2000 Stefan Wedner
//
// This program/module is free software for non-commercial use.
//
//
//============================================================================
//----------------------------------------------------------------------------
//
// File:     peano.h
// See also: peano.C, peano.tpl, node.h, node.C
// Purpose:  computing the peano constants of quadrature formulas
//     
// ===========================================================================


#include <string>
#include <fstream>
#include "imath.hpp"
#include "l_interval.hpp"
#include "node.h"

#define odd(integer) (integer%2)


// ----------------------------------------------------------------------------
// auxiliary functions
// ----------------------------------------------------------------------------

template <class T> inline T truncated_diff(const T& x, const T& t ){
// return the positive part (x-t)_+ of the interval x - t
  T y = x - t;          
  if( Sup(y)<=0 )
    return T(0.0,0.0);  
  if( Inf(y)<0 )
    return T(0,Sup(y)); 
  return y;                     
}


/* inline bool intersecting_nonempty(const interval& a, const interval& b ){ */
/*   return Sup(a)>=Inf(b) && Inf(a)<=Sup(b); */
/* } */


// ----------------------------------------------------------------------------
// ------------------------ readdata_symm, readdata ---------------------------
// ----------------------------------------------------------------------------


template<class T> void readdata(ifstream& external_data, int number_nodes, 
                    T* node1, T* weight1, interval* node2, interval* weight2){
 // reading nodes from file
  int i;
  string tmp;

  for(i = 1; i <= number_nodes; i++){
     external_data >> tmp;
     tmp >> node1[i];
     node2[i] = node1[i];
  }

 // reading weights from file
  for(i = 1; i <= number_nodes; i++){
     external_data >> tmp;
     tmp >> weight1[i];
     weight2[i] = weight1[i]; 
  }
}


template<class T> void readdata_symm(ifstream& external_data, int n, T* node1,
                              T* weight1, interval* node2, interval* weight2){
  // reading nodes and weights from file   
  int i;
  string tmp;

  for(i = 1; i <= n / 2; i++){
    external_data >> tmp;
    tmp >> node1[i];
    node2[i] = node1[i];
    node1[n - i + 1] = - node1[i];
    node2[n - i + 1] = - node1[i];
  }

  if(n % 2){
    external_data >> tmp;
    tmp >> node1[n / 2 + 1];
    node2[n / 2 + 1] =  node1[n / 2 + 1];
  }   

  for(i = 1; i <= n / 2; i++){  
    external_data >> tmp;
    tmp >> weight1[i];
    weight2[i] = weight2[n - i + 1] = weight1[i];
    weight1[n - i + 1] = weight1[i];
   
  }  

  if(n % 2){
    external_data >> tmp;
    tmp >> weight1[n / 2 + 1];
      weight2[n / 2 + 1] = weight1[n / 2 + 1];
  }   
}



// ----------------------------------------------------------------------------
// class peano_base (abstract class)
// ----------------------------------------------------------------------------
//
// pure  virtual function
//
//   virtual interval kernel(interval x);
//   virtual interval derivation1(interval x);
//   virtual interval derivation2(interval x);
//   virtual void integrate_stamm(real from, real to);
//   virtual void use_symmetrie();


class peano_base{

 public: 
   
  peano_base(string filename, bool symm = true);
  peano_base();
  // initialisation of : node, weight, symmetric, number_nodes, positive_part
  //                     negative_part,  
  virtual ~peano_base(){ delete[] node; delete[] weight;}

  void calculate(int);

  friend ostream& operator<<(ostream&, const peano_base&);
  int numbernodes(){ return number_nodes;  }
 protected: 

   interval *node, *weight;  
   int  order_kernel;
   int  number_nodes;
   bool symmetric;

   idotprecision positive_part, negative_part;
   real epsilon;
 
   static idotprecision i_dot;
   static int start_index;
   static real integrate_from, integrate_to;
   
   interval mvf_kernel(interval x);
   interval mvf_derivation1(interval x);
   void extented_newton(interval x);
   void check_closer();
   void integrate_rectangle(const interval& x, const interval& K_x);

 private:
   virtual interval kernel(int k) = 0;
   // enclosure of (s-1)! K_s([x_k]) 
   virtual interval kernel(interval x) = 0;
   // enclosure of (s-1)! K_s(x)    x in [x_start_index, x_start_index+1]
   virtual interval derivation1(interval x) = 0;
   // enclosure of (s-1)! K_s(x)   x in [x_start_index, x_start_index+1]
   virtual interval derivation2(interval x) = 0;
   // enclosure of (s-1)! K_s(x)  x in [x_start_index, x_start_index+1]
   virtual void integrate_stamm(real from, real to) = 0;
   // add integral of s! K_s(x)  x in [x_start_index, x_start_index+1]
   virtual void use_symmetrie(){}
   // if K is nonsymmetric this function is empty
   virtual void init_data(){}
   // only for m.p.v. kernels or weighted kernels
};



// ----------------------------------------------------------------------------
// class peano (legendre w = 1, [-1,1])
// ----------------------------------------------------------------------------


class peano: public peano_base{
 public:
  peano(string filename, real eps = 1e-6):peano_base(filename){
                                                               epsilon = eps;} 
 private:
 
  interval kernel(interval t);
  interval kernel(int k);
  interval derivation1(interval t);
  interval derivation2(interval t);
  void integrate_stamm(real from, real to); 
  void use_symmetrie();
};



// ----------------------------------------------------------------------------
// class peano_weight (w = 1/sqrt(x), [0, 1])
// ----------------------------------------------------------------------------



template<class T = l_interval> class peano_weight: public peano_base{

 public:
  peano_weight(string, int = 1, real  = 1e-6);
 
  ~peano_weight(); 

  T high();

 private:
  T *node, *weight;
  T** s_over_nu;
  T gamma; 
  virtual T kernel_alpha(real, int);
  interval kernel(int k);
  interval kernel(interval t);
  interval derivation1(interval t);
  interval derivation2(interval t);
  void integrate_stamm(real from, real to); 
  void init_data();
 
};



// ----------------------------------------------------------------------------
// class peano_cauchy 
// ----------------------------------------------------------------------------


template<class T = interval> class peano_cauchy: public peano_base{

 public: 
  peano_cauchy(string, int = 1, real  = 1e-6);
 
  ~peano_cauchy();

 private:
  T *node, *weight;
  T*  sum_nu ; // sum_nu[0] initialized with  0!!!!!!
  T** s_over_nu; 

  T kernel_beta(real x, int s);
  interval kernel(int k);
  interval kernel(interval t);
  interval derivation1(interval t);
  interval derivation2(interval t);
  void integrate_stamm(real from, real to); 
  void use_symmetrie();
  void init_data();
};



