//============================================================================
//
//                              Program/Module
//                                   from
//                C++ package for verified numerical integration
//                  of singular and strong singular functions
//                      Copyright (c) 2000 Stefan Wedner
//
// This program/module is free software for non-commercial use.
//
//
//============================================================================
//----------------------------------------------------------------------------
//
// File:     quadrature.h
// See also: quadrature.C
// Purpose:  header (quadrature.C)
//     
// ===========================================================================


#include <algorithm>
#include <iterator>
#include <iostream>
#include "interval.hpp"
#include "./taylor/f_ari.h"

class integral;

// ----------------------------------------------------------------------------
// --- class base_interval ----------------------------------------------------
// ----------------------------------------------------------------------------

class base_interval{
 public:

  base_interval(real a, real b, int k, int l, integral& r_i): inf(a), sup(b),
             m_index(k), n_index(l), quad(false), ri(r_i), dremainder(0),
             r_inf(0), r_sup(0){}
  // base_interval(){ cout << " base_interval empty-constr. " << endl;}
  //  virtual ~base_interval();

  real dremainder;

 protected:
  real inf, sup;
  real r_inf, r_sup;
  interval quadsum;

  int  n_index, m_index, n, m;
  bool quad;
  integral& ri;

  static dotprecision r_dot;

  virtual void add_quadraturesum() = 0;
  virtual void subdivide() = 0;
  virtual void print_interval(ostream& os){ 
        os  << " intervaltype: unknown " << *this << endl;}

   // base_interval(const base_interval& t);
  //  base_interval& operator=(const base_interval&);

  friend ostream& operator<<(ostream& os, const base_interval& out);
  friend istream& operator>>(istream& is, base_interval& in);

  friend class integral; 
};



// ----------------------------------------------------------------------------
// --- class pbase_interval ---------------------------------------------------
// ----------------------------------------------------------------------------


class pbase_interval{
 
 public:
  pbase_interval(){ pointer = 0;}
  pbase_interval(base_interval& r_i){ pointer = &r_i; }
  bool operator<(const pbase_interval&) const;
  bool operator>(const pbase_interval&) const;
  bool operator==(const pbase_interval&) const;
  friend ostream& operator<<(ostream& os, const pbase_interval& out);
  friend istream& operator>>(istream& is, pbase_interval& in);
  base_interval* pointer;
};


// ----------------------------------------------------------------------------
// --- class riemann ----------------------------------------------------------
// ----------------------------------------------------------------------------


class riemann:public base_interval{
 
 public:
  // riemann(real d = 0){ dremainder = d; } // remove
  riemann(real a, real b, int k, int l, integral& r_i):
                                       base_interval(a, b, k, l, r_i){
    m = m_index * m_stride + m_min;
    n = n_index * n_stride + n_min;
  }
  //   riemann(real, real, int, int, integral&,)
  void set_riemann(interval, int, int);
  void set_riemann(int, int, interval);

  static interval **node, **weight, **peano; // quadrature formulas
  static const int n_min, n_max, n_dim; 
  static const int m_min, m_max, m_dim;
  static const int m_stride, n_stride;

  void add_quadraturesum();
  void subdivide();
  void choose_qf(real eps_local);
  void print_interval(ostream& os){
    os << " intervaltype: riemann " << *this << endl;
  } ;  

  static interval** init_data();
  static void free_data();

};

// ----------------------------------------------------------------------------
// --- class integral ---------------------------------------------------------
// ----------------------------------------------------------------------------

class integral{

public:
  integral();
  integral(integrand&);
  ~integral();
  interval integrate(real epsabs);
  interval integrate(real a, real b, real epsabs);
  interval integrate(interval a, real b, real epsabs);
  interval integrate(real a, interval b, real epsabs);
  interval integrate(interval a, interval b, real epsabs);
  friend ostream& operator<<(ostream& os, const integral& out);

  void gaa();

  void print_division();
  void remainders(real a, real b);

protected:

  real inf, sup;
  integrand f;  
  real eps_abs, eps_rel;
 

  dotprecision Rinf, Rsup;
  idotprecision Quadsum;

  static const int max_heapsize;  // heap: administration of subintervals 
  int fcounter;

  int heapsize;
  pbase_interval* heap;
  // functions:
  base_interval* pop();
  void push(base_interval& r_i);
  void print_heap();
  
  class bound_error;
 private: 
  virtual void start_division();

 friend class riemann;

};


// inline memberfunctions:
// heap administration: push, pop, print_heap

inline base_interval* integral::pop(){
    pop_heap(heap, heap + heapsize--);
    accumulate(Rinf, -1.0, heap[heapsize].pointer->r_inf);
    accumulate(Rsup, -1.0, heap[heapsize].pointer->r_sup);
    return heap[heapsize].pointer;
}

inline void integral::push(base_interval& ri){
    heap[heapsize] = *new pbase_interval(ri);
    push_heap(heap, heap + ++heapsize); 
    accumulate(Rinf, 1.0, ri.r_inf);
    accumulate(Rsup, 1.0, ri.r_sup);
}

inline void integral::print_heap(){
    copy(heap, heap + heapsize, ostream_iterator<pbase_interval>(cout, "  "));
}



// ----------------------------------------------------------------------------
// --- class sqrt_integral ----------------------------------------------------
// ----------------------------------------------------------------------------

class sqrt_integral: public integral{
 public:
  sqrt_integral(integrand& g);
  ~sqrt_integral();
 protected:
  integrand f_w; 
  
 private: 
  void start_division();
  static int sqrt_counter;
  friend class sqrt_interval;
};



// ----------------------------------------------------------------------------
// --- class cauchy_integral --------------------------------------------------
// ----------------------------------------------------------------------------


class cauchy_integral: public integral{
 public:
  cauchy_integral();
  cauchy_integral(integrand& g, real lambda);

 protected:
  real lambda; // collocation point!  
  integrand f_c; 
  
 private: 
  void start_division();

 friend class cauchy;
};



// ----------------------------------------------------------------------------
// --- class sqrt_interval ----------------------------------------------------
// ----------------------------------------------------------------------------


class sqrt_interval: public base_interval{

public:
  
  //  sqrt_interval();
  //  sqrt_interval(real r){ dremainder = r;}
  sqrt_interval(real a, real b, int k, int l, sqrt_integral& sq):
                                       base_interval(a, b, k, l, sq){
    m = m_index * n_stride + n_min;
    n = n_index * n_stride + n_min;
  }

  // nodes, weights and peanoconstants
  static interval **peano;
  static interval **node;
  static interval **weight;                               
  
  static const int n_min, n_max, n_stride, n_dim;

  static interval** init_data();
  static void free_data();

  void add_quadraturesum();
  void choose_qf(const real&);
  void subdivide();  

  void set_sqrt(interval, int, int);
  void set_sqrt(int, int, interval);

  void print_interval(ostream& os){ 
          os  << " intervaltype: sqrt " << *this << endl;}
};



// ----------------------------------------------------------------------------
// --- class cauchy -----------------------------------------------------------
// ----------------------------------------------------------------------------


class cauchy: public base_interval{
public:
  // cauchy(){ cout << " cauchy const " << endl; }
  // cauchy(real r){ dremainder = r;  cout << " cauchy const real " << endl;}
  cauchy(real a, real b, int k, int l, cauchy_integral& c_i):
                                       base_interval(a, b, k, l, c_i){
    m = m_index * cauchy::m_stride  + cauchy::m_min;
    n = n_index * riemann::n_stride + riemann::n_min;
  }
  static interval **peano;                               
  static const int m_min, m_max, m_stride, m_dim;

  static interval** init_data();
  static void free_data();


  void add_quadraturesum();
  virtual void choose_qf(const real&);
  void subdivide();  

  void set_cauchy(interval, int, int);
  void set_cauchy(int, int, interval);

   void print_interval(ostream& os){ 
          os  << " intervaltype: cauchy " << *this << endl;}
};


// ----------------------------------------------------------------------------
// --- class hadamard_integral ------------------------------------------------
// ----------------------------------------------------------------------------


class hadamard_integral: public cauchy_integral{
 public:
  hadamard_integral(integrand& g, real lambda);

 private: 
  void start_division();
  friend class hadamard;
};


// ----------------------------------------------------------------------------
// --- class hadamard ---------------------------------------------------------
// ----------------------------------------------------------------------------


class hadamard: public cauchy{
public:
  hadamard(real a, real b, int k, int l, hadamard_integral& h_i):
                                                 cauchy(a, b, k, l, h_i){
  }
  hadamard(real a, real b, hadamard_integral& h_i):cauchy(a, b, 0, 0, h_i){}
  static int const had_max; 
  static interval* peano; 
  void add_quadraturesum();
  void choose_qf(const real&);
  void set_hadamard(interval r, int k, int l);
  void print_interval(ostream& os){ 
          os  << " intervaltype: hadamard " << *this << endl;}
  static interval* init_data();
};


// ----------------------------------------------------------------------------
// --- errorhandling
// ----------------------------------------------------------------------------

class integral::bound_error{
 public:
  bound_error(){}
};

extern operand x, y;
