//============================================================================
//
//                              Program/Module
//                               from CLAVIS
//                C++ library for verified numerical integration
//                  of singular and strong singular functions
//                     Copyright (c) 2000 Stefan Wedner
//
//          This program/module is free software for non-commercial use. 
//
// 
//============================================================================
//----------------------------------------------------------------------------
//
// File:    f_ari.h
// Purpose: taylorarithmetic (one dimensional and two dimensional)
//
//============================================================================
// Known restrictions: 
//      
//    Avoid assignment:   
//      operand r,s,t:
//       r = s;
//       s = t; 
//
//
//============================================================================

#include<interval.hpp>
#include<idot.hpp>
#include<rmath.hpp>
#include<imath.hpp>
#include<string>

using namespace cxsc;
using namespace std;

// #define MEMORY_CHECK

#ifdef  MEMORY_CHECK
int testdest();            // Ueberpruefen des Speichermanagements
#endif

const int max_named = 5;   // Maximalzahl der benannten Operanden 
                           // benannt = globale oder lokale Variable                           

typedef interval (*i_function)(const interval& in); 


// ---------------------------------------------------------------------------
// class operand 
// ---------------------------------------------------------------------------

class operand{

 public:

  enum operand_type{ Uminus, Plus, Minus, Mult, Div, Var, Const, Exp, Ln, 
                     Sin, Cos, Tan, Cot, Sinh, Cosh, Tanh, Coth,
                     Asin, Acos, Atan, Acot, Asinh, Acosh, Atanh, Acoth,
                     Sqr, Sqrt, Pow, Power, Inc, Var_1, Var_2 };

  explicit operand(operand_type t_var = Var_1); 
  explicit operand(operand& ri);
  ~operand(); //
  // destruktor loescht Knoten mit del = true und alle Kinder (rekursiv)  
   
  void operator=(operand&);
  // kopiert Wert des rechten Kopfes nach links 
  // rechter Knoten (und nur der) wird geloescht falls del = true

  template<class scalar> void operator=(const scalar& ri);

  void transpose();
  static void set_detail(bool = true); // detailliertere Ausgabe

 private:  

  class error_foperand; 

   explicit operand(const real& constant);
   explicit operand(const interval& constant);
   operand (operand_type otype, operand& Left, operand& Right);
   operand (operand_type otype, operand& Left); 
   operand (operand_type otype, bool deps, bool dept);
   operand (const operand* op, operand* Left, operand* Right);
   // Spezialfall: del wird abhaengig auf op->del gesetzt. 
   // wird bei cp_operand verwendet.
 
   operand_type o_type; 
   operand *left, *right;
   bool dep_t;  // operand abhaengig von Variable t?
   bool dep_s;  // operand abhaengig von Variable s?
   bool status; // s.u.
   interval value;  
   int exponent;      
   union{
     interval *tk;    
     interval **TK;   
   };
   bool del;  // del = false bedeutet Operand ist ein benanntes Objekt
              // oder eine Variable. Ist der Operand Teil eines Integranden
              // dann kann auf diesen Knoten mehrfach verwiesen werden.

   // static variables

   static bool detail;
   static bool done;
   static bool triangular;

   static int dim_var1;
   static int dim_var2;

   static int s_lb;
   static int s_ub;
   static int t_lb;
   static int t_ub;
   static idotprecision i_accu;
   
   static operand* var1;
   static operand* var2;
   static operand* named[2 * max_named];
   static int index_named;
   // internal functions

   void set_variables(operand*, operand*);
   void set_value();
   void diff();
   void trigo1(i_function stf);
   void trigo2(i_function stf);   

   // memory management

   void delete_operand(); // delete operand (not recursive)

   void save_named();
   void free_taylor(); 
   void alloc_taylor();
   void alloc_rectangle();
   void alloc_triangle();

  // friend declarations 

  friend class integrand;
  friend string   o_type2string(operand_type op_type);
  friend operand* cp_operand(operand* ); // copy ftree

  friend operand& exp(operand& Left); 
  friend operand& ln(operand& Left);
  friend operand& sqr(operand& Left);
  friend operand& sqrt(operand& Left);
  friend operand& pow(operand& Left, const real& Right);
  friend operand& power(operand& Left, int Right);
  friend operand& sin(operand& Left);
  friend operand& cos(operand& Left);
  friend operand& tan(operand& Left);
  friend operand& cot(operand& Left);
  friend operand& asin(operand& Left);
  friend operand& acos(operand& Left);
  friend operand& atan(operand& Left);
  friend operand& acot(operand& Left); 
  friend operand& sinh(operand& Left);
  friend operand& cosh(operand& Left);
  friend operand& tanh(operand& Left);
  friend operand& coth(operand& Left);
  friend operand& asinh(operand& Left);
  friend operand& acosh(operand& Left);
  friend operand& atanh(operand& Left);
  friend operand& acoth(operand& Left);
  friend ostream& operator<<(ostream& os, const operand& out);
  friend operand& operator-(operand& Left);
  friend operand& operator+(operand&, operand&);
  friend operand& operator-(operand&, operand&);
  friend operand& operator*(operand&, operand&);
  friend operand& operator/(operand&, operand&);

  template<class scalar> friend operand& operator+(operand& Left, 
                                                   const scalar& r);
  template<class scalar> friend operand& operator+(const scalar& r, 
                                                   operand& Ri);
  template<class scalar> friend operand& operator-(operand& Left,
                                                   const scalar& r);
  template<class scalar> friend operand& operator-(const scalar& r, 
                                                   operand& Ri);
  template<class scalar> friend operand& operator*(operand& Left,
                                                   const scalar&r);
  template<class scalar> friend operand& operator*(const scalar& r, 
                                                    operand& Ri);
  template<class scalar> friend operand& operator/(operand& Left,
                                                   const scalar& r);
  template<class scalar> friend operand& operator/(const scalar& r, 
                                                   operand& Ri);
};



// ----------------------------------------------------------------------------
// functions  exp, ln, sqr, sqrt, pow , ...........
// ----------------------------------------------------------------------------

operand& exp(operand& Left); 
operand& ln(operand& Left);
operand& sqr(operand& Left);
operand& sqrt(operand& Left);
operand& pow(operand& Left, const real& Right);
operand& power(operand& Left, int Right);
operand& sin(operand& Left); 
operand& cos(operand& Left); 
operand& tan(operand& Left); 
operand& cot(operand& Left); 
operand& asin(operand& Left);
operand& acos(operand& Left);
operand& atan(operand& Left);
operand& acot(operand& Left);
operand& sinh(operand& Left);
operand& cosh(operand& Left);
operand& tanh(operand& Left);
operand& coth(operand& Left);
operand& asinh(operand& Left);
operand& acosh(operand& Left);
operand& atanh(operand& Left);
operand& acoth(operand& Left);


// ----------------------------------------------------------------------------
// operators -,+,-,*,/
// ----------------------------------------------------------------------------

operand& operator-(operand& Left);

operand& operator+(operand&, operand&);

template<class scalar>operand& operator+(operand& Left, const scalar& r){
   operand* tmp = new operand(r);
   return Left + *tmp;  
}

template<class scalar> operand& operator+(const scalar& r, operand& Ri){
   operand* tmp = new operand(r);
   return *tmp + Ri;  
}

operand& operator-(operand&, operand&);

template<class scalar> operand& operator-(operand& Left,const scalar& r){
   operand* tmp = new operand(r);
   return Left - *tmp;  
}

template<class scalar>  operand& operator-(const scalar& r, operand& Ri){
   operand* tmp = new operand(r);
   return *tmp - Ri;  
}

operand& operator*( operand&, operand&);

template<class scalar> operand& operator*( operand& Left,const scalar&r){
   operand* tmp = new operand(r);
   return Left * *tmp;  
}

template<class scalar> operand& operator*(const scalar& r, operand& Ri){
   operand* tmp = new operand(r);
   return *tmp * Ri;  
}

operand& operator/( operand&,  operand&);

template<class scalar> operand& operator/( operand& Left,const scalar& r){
   operand* tmp = new operand(r);
   return Left / *tmp;  
}

template<class scalar> operand& operator/(const scalar& r, operand& Ri){
    operand* tmp = new operand(r);
    return *tmp / Ri;  
}

template<class scalar> void operand::operator=(const scalar& ri){
  o_type = Const;
  left   = 0;
  right  = 0;
  value  = ri;
  dep_s  = false;
  dep_t  = false;
  status = false;
  del    = false;
  tk     = 0;
}

// ---------------------------------------------------------------------------
// class integrand
// ---------------------------------------------------------------------------

class integrand{
 public:

  integrand(int dim1 = 0, int dim2 = 0);
  integrand(const integrand&); 
  //  integrand(operand& right);  
  ~integrand();
  
  interval value(const interval& s);
  interval value(const interval&, const interval&) const;
  interval taylor(int) const;
  interval taylor(int, const interval&, const interval& = _interval(0)) const;
  interval taylor(int, int) const;
  interval taylor(int, int, const interval&, const interval&) const;

  integrand& operator=(operand&); 
  integrand& operator=(integrand&);
  template<class scalar> integrand& operator=(const scalar& ri);

  integrand& operator+(operand&);
  integrand& operator-(operand&);
  integrand& operator*(operand&);
  integrand& operator/(operand&);

  integrand& operator+(integrand& right);
  integrand& operator-(integrand& right);
  integrand& operator*(integrand& right);
  integrand& operator/(integrand& right);

  integrand& transpose();  
  friend ostream& operator<<(ostream&, const integrand&);

  void set_dim(int = 0, int = 0);

  class error;

 private:

  int dim_var1;
  int dim_var2;
 
  mutable int diff_s;
  mutable int diff_t;

  bool triangular;
  bool dep_s;
  bool dep_t;
  mutable bool status;
  bool del;

  operand* var1;
  operand* var2;
  operand* ftree; 
  
  void alloc_taylor(int dim1 = 0, int dim2 = 0);
  void free_taylor();
  void delete_ftree(); // free ftree
  void set_var1(const interval& ,bool = true) const;
  void set_var2(const interval& ,bool = true) const;
  void set_var();

};


template<class scalar> integrand& integrand::operator=(const scalar& ri){
  // free memory left side
  if (ftree) free_taylor();
  delete_ftree();
  // allocate memory for variables
  set_var();
  dep_s = false;
  dep_t = false;
  ftree = new operand(ri);
  alloc_taylor();
  return *this;
}


class operand::error_foperand{
 public:
  operand_type o_type;
  error_foperand(operand_type i = Inc){ o_type = i;}
};


class integrand::error{
 public:
  int i;
  explicit error(int j){ i = j;}
};

